/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2008 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.auxml;

import java.io.*;
import java.util.*;
import java.text.SimpleDateFormat;
import org.w3c.dom.*;

/** 
 * Represents all information about one Module 
 */
public class Module extends Object {

    private Properties moduleProperties = new Properties();
    private Properties manifestProperties = null;
    private Map localized_manifestProperties = new HashMap();
    private Properties l10nProperties = null;
    private ExternalPackage external_package = null;
    private String description = null;
    private String notification = null;
    private String nbmName = null;
    private String language = "en";
    private String os = null;
    private List dependencies = new ArrayList();
    private Set moduleDeps = new HashSet();
    private Set versionDeps = new HashSet();
    private boolean isNew = false;
    private boolean isUpdated = false;
    private ModuleGroup parentGroup = null;
    public static final String OS_NAME = "os_name";
    public static final String DISTRIBUTION_BASE = "distribution_base_url";
    public static final String DISTRIBUTION_DIR = "distribution_dir";
    public static final String IMPL_VERSION = "impl-version";
    public static final String LICENSE = "license";
    public static final String HOMEPAGE = "homepage";
    public static final String CODENAMEBASE = "codenamebase";
    public static final String MODULE_NOTIFICATION = "notification_file";
    public static final String ADD_DEPS = "add_deps";
    public static final String DISTRIBUTION_SUFFIX = "distribution_url_suffix";
    public static final String DISTRIBUTION_URL = "distribution_url";

    /**
     * Creates Module object from NBMfile object
     */
    public Module(NBMfile nbmFile, ModuleGroup parentGroup) {
        this.parentGroup = parentGroup;
        nbmName = nbmFile.getName();
        os = nbmFile.os;
        Document info_xml_doc;
        List supported_langs = nbmFile.getSupportedLangs();
        String one_lang;
        for (Iterator i = supported_langs.iterator(); i.hasNext();) {
            one_lang = (String) i.next();
            info_xml_doc = XMLutils.openXML(nbmFile.getInfoXMLStream(one_lang));
            loadFromDocument(info_xml_doc);

            Properties localized_manifest = new Properties(manifestProperties);
            localized_manifestProperties.put(one_lang, localized_manifest);
        }
        info_xml_doc = XMLutils.openXML(nbmFile.getInfoXMLStream());
        loadFromDocument(info_xml_doc);
        setConfigData(nbmFile);
        setNbmSize(nbmFile.getSize());
        if (isModule()) {
            selectDescription();

            System.out.print(AppSetup.odsadenie);
            System.out.println("|- Module nbm: " + nbmName + " " + getNameSpecVersion() + " " + isNew);
        }
        if (isL10N()) {
            System.out.print(AppSetup.odsadenie);
            System.out.println("|- L10N nbm: " + nbmName + " " + getNameSpecVersion());
        }
        applyModulesSetup();
    }

    /**
     * Returns module name
     */
    public String getName() {
        if (isModule()) {
            return (manifestProperties.getProperty("OpenIDE-Module"));
        } else {
            return (moduleProperties.getProperty(CODENAMEBASE));
        }
    }

    /**
     * Returns module codebasename
     */
    public String getCodebaseName() {
        return (moduleProperties.getProperty(CODENAMEBASE));

    }

    /**
     * Returns module specification version
     */
    public String getSpecVersion() {
        if (isModule()) {
            return (manifestProperties.getProperty("OpenIDE-Module-Specification-Version"));
        } else if (isL10N()) {
            return (l10nProperties.getProperty("module_spec_version"));
        } else {
            return ("*");
        }
    }

    /**
     * Returns module license name
     */
    public String getLicense() {
        return (moduleProperties.getProperty("license"));
    }

    /**
     * Returns module license name
     */
    public String getDistURL() {
        return (moduleProperties.getProperty("distribution"));
    }
    
    /**
     * Returns string "name spec_version"
     */
    public String getNameSpecVersion() {
        return (getName() + " " + getSpecVersion());
    }

    /**
     * Returns module NBM size
     */
    public String getNbmSize() {
        return (moduleProperties.getProperty("downloadsize"));
    }

    /**
     * Sets module NBM size
     */
    public void setNbmSize(long size) {
        moduleProperties.setProperty("downloadsize", "" + size);
    }

    /**
     * Sets module new implementation version
     */
    public void setImplVersion() {
        SimpleDateFormat formatter = new SimpleDateFormat("yyMMddHHmmss");
        manifestProperties.setProperty("OpenIDE-Module-Implementation-Version", formatter.format(new Date()));
    }

    /**
     * Returns set of module dependencies 
     */
    public Set getModuleDeps() {
        return (moduleDeps);
    }

    /**
     * Returns set of module dependencies with concrete version
     */
    public Set getVersionDeps() {
        return (versionDeps);
    }

    /**
     * Returns parent module group
     */
    public ModuleGroup getParentGroup() {
        return (parentGroup);
    }

    /**
     * Returns YES if it is module
     */
    public boolean isModule() {
        return (manifestProperties != null);
    }

    /**
     * Returns YES if it is L10N package
     */
    public boolean isL10N() {
        return (l10nProperties != null);
    }

    /**
     * Sets is_updated flag
     */
    public void setUpdated() {
        isUpdated = true;
    }

    public boolean isUpdated() {
        return (isUpdated);
    }
    
    public boolean isStaging() {
        return getParentGroup().isStaging();
    }

    /**
     * Selects which Module Description to use:
     * there are 2 possible ways how to store Module Description in info.xml
     * in tag <Description> or in manifest property "OpenIDE-Module-Long-Description"
     * If <Description> is missing then property "OpenIDE-Module-Long-Description"
     * is used.
     */
    private void selectDescription() {

        if (description == null) {
            description = manifestProperties.getProperty("OpenIDE-Module-Long-Description");
        }

        if (description == null) {
            description = "";
        }

    }

    /**
     * Test if the module should be created or updated
     */
    private boolean isNew() {
        return (!AppSetup.loaded_modules.contains(getName()));
    }

    /**
     * Stores module information to XML element 
     * input:
     * resDoc - target document
     * parElement - parent element
     */
    public void toCatalogElement(Element parElement) {
        if (ModulesCatalog.staging && isUpdated) {
            System.out.println(" module was updated");
            return;
        }

        Document resDoc = parElement.getOwnerDocument();
        isNew = isNew();

        if (isNew) {
            AppSetup.loaded_modules.add(getName());
        }
        if (isModule()) {
            System.out.println(" |- saving module " + getName());
        } else if (isL10N()) {
            System.out.println(" |- saving l10n package " + getName());
        }
        Element module_element = resDoc.createElement("module");
        String prop_name;
        for (Enumeration e = moduleProperties.propertyNames(); e.hasMoreElements();) {
            prop_name = (String) e.nextElement();
            module_element.setAttribute(prop_name, moduleProperties.getProperty(prop_name));
        }

        /* parameter release is obsolete        
        String release = NB_AU_info.getDistribution( name_spec_version() );
        if (release==null) release = "auto";
        module_element.setAttribute("release", release);
         */
        parElement.appendChild(module_element);

        if (description != null) {
            Element desc_element = resDoc.createElement("description");
            module_element.appendChild(desc_element);

            CDATASection desc_cdata_section = resDoc.createCDATASection(description);
            desc_element.appendChild(desc_cdata_section);
        }

        if (external_package != null) {
            external_package.toCatalogElement(module_element);
        }
        if (notification != null) {
            Element notif_element = resDoc.createElement("module_notification");
            module_element.appendChild(notif_element);

            CDATASection notif_cdata_section = resDoc.createCDATASection(notification);
            notif_element.appendChild(notif_cdata_section);
        }

        Properties act_manifestProperties = null;

        if (ModulesCatalog.language != null) {
            act_manifestProperties = (Properties) localized_manifestProperties.get(ModulesCatalog.language);
        }

        if (act_manifestProperties == null) {
            act_manifestProperties = manifestProperties;
        }


        if (act_manifestProperties != null) {
            Element manifest_element = resDoc.createElement("manifest");
            for (Enumeration e = act_manifestProperties.propertyNames(); e.hasMoreElements();) {
                prop_name = (String) e.nextElement();
                manifest_element.setAttribute(prop_name, act_manifestProperties.getProperty(prop_name));
            }
            module_element.appendChild(manifest_element);
        }


        if (l10nProperties != null) {
            Element l10n_element = resDoc.createElement("l10n");
            for (Enumeration e = l10nProperties.propertyNames(); e.hasMoreElements();) {
                prop_name = (String) e.nextElement();
                l10n_element.setAttribute(prop_name, l10nProperties.getProperty(prop_name));
            }
            module_element.appendChild(l10n_element);
        }

        addExtraInfo(module_element);
    }

    @Override
    public String toString() {
        return getGroupPath() + "::" + getName();
    }
    
    public String getGroupPath() {
        StringBuilder sb = new StringBuilder();
        ModuleGroup group = getParentGroup();
        while (group != null) {
            sb.insert(0, group.getName());
            group = group.getParentGroup();
            if (group != null) {
                sb.insert(0, "/");
            }
        }
        // root group
        if (sb.length() == 0) {
            sb.append("/");
        }
        return sb.toString();
    }

    /**
     * Stores additional module information (information not stored in info.xml), 
     * to <extra_info> element 
     * input:
     * resDoc - target document
     * parElement - parent element
     */
    private void addExtraInfo(Element parElement) {
        Document resDoc = parElement.getOwnerDocument();

        Element extra_info_element = resDoc.createElement("extra_info");
        parElement.appendChild(extra_info_element);

        Element request_element = resDoc.createElement("request");
        extra_info_element.appendChild(request_element);
        if (isNew) {
            request_element.setAttribute("type", "create");
        } else {
            request_element.setAttribute("type", "update");
        }
        Element module_name_element = resDoc.createElement("module_name");
        extra_info_element.appendChild(module_name_element);

        Text module_name_text = resDoc.createTextNode(nbmName);
        module_name_element.appendChild(module_name_text);

        /*        
        Element module_name_element = resDoc.createElement("module_name");
        extra_info_element.appendChild(module_name_element);
         */

        Dependency one_dep;
        for (Iterator i = dependencies.listIterator(); i.hasNext();) {

            one_dep = (Dependency) i.next();

            Element dep_element = resDoc.createElement(one_dep.type);
            extra_info_element.appendChild(dep_element);

            Element dep_module = resDoc.createElement("dependency_module_name");
            dep_element.appendChild(dep_module);

            Text dep_text = resDoc.createTextNode(one_dep.module);
            dep_module.appendChild(dep_text);

            if (one_dep.version != null) {
                Element dep_version = resDoc.createElement("dependency_version");
                dep_version.setAttribute("condition", one_dep.condition);
                dep_element.appendChild(dep_version);

                dep_text = resDoc.createTextNode(one_dep.version);
                dep_version.appendChild(dep_text);
            }

        }

    }

    /**
     * Sets module info from XML
     */
    private void loadFromDocument(Document infoXMLDoc) {
        Element element;
        element = infoXMLDoc.getDocumentElement();

        if ((element != null) && element.getTagName().equals("module")) {
            scanElementModule(element);
        }
    }

    /**
     * Scans module element
     */
    private void scanElementModule(Element element) { // <module>
        NamedNodeMap attrs = element.getAttributes();
        for (int i = 0; i < attrs.getLength(); i++) {
            Attr attr = (Attr) attrs.item(i);
            moduleProperties.setProperty(attr.getName(), attr.getValue());
        }
        NodeList nodes = element.getChildNodes();
        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            switch (node.getNodeType()) {
                case Node.ELEMENT_NODE:
                    Element nodeElement = (Element) node;
                    if (nodeElement.getTagName().equals("manifest")) {
                        scanElementManifest(nodeElement);
                    }
                    if (nodeElement.getTagName().equals("l10n")) {
                        scanElementL10N(nodeElement);
                    }
                    if (nodeElement.getTagName().equals("description")) {
                        scanElementDescription(nodeElement);
                    }
                    if (nodeElement.getTagName().equals("module_notification")) {
                        scanElementNotification(nodeElement);
                    }
                    if (nodeElement.getTagName().equals("external_package")) {
                        external_package = new ExternalPackage(nodeElement);
                    }
                    if (nodeElement.getTagName().equals("license")) {
                        scanElementLicense(nodeElement);
                    }
                    break;
            }
        }
    }

    /**
     * Scans manifest element
     */
    private void scanElementManifest(Element element) { // <manifest>
        manifestProperties = new Properties();

        NamedNodeMap attrs = element.getAttributes();
        for (int i = 0; i < attrs.getLength(); i++) {
            Attr attr = (Attr) attrs.item(i);
            if (isValidManifestAttribute(attr.getName())) {
                manifestProperties.setProperty(attr.getName(), attr.getValue());
            }
        }

        String dep_string = manifestProperties.getProperty("OpenIDE-Module-Package-Dependencies");
        if (dep_string != null) {
            readDependencies(dep_string, "package_dependency");
        }
        dep_string = manifestProperties.getProperty("OpenIDE-Module-Module-Dependencies");
        if (dep_string != null) {
            readDependencies(dep_string, "module_dependency");//    printModuleDeps();
        }
    }

    /**
     * validate manifest attributs
     */
    private boolean isValidManifestAttribute(String attrName) {
        if (attrName.equals("OpenIDE-Module")) {
            return (true);
        }
        if (attrName.equals("OpenIDE-Module-Name")) {
            return (true);
        }
        if (attrName.equals("OpenIDE-Module-Specification-Version")) {
            return (true);
        }
        if (attrName.equals("OpenIDE-Module-Implementation-Version")) {
            return (true);
        }
        if (attrName.equals("OpenIDE-Module-Module-Dependencies")) {
            return (true);
        }
        if (attrName.equals("OpenIDE-Module-Package-Dependencies")) {
            return (true);
        }
        if (attrName.equals("OpenIDE-Module-Java-Dependencies")) {
            return (true);
        }
        if (attrName.equals("OpenIDE-Module-IDE-Dependencies")) {
            return (true);
        }
        if (attrName.equals("OpenIDE-Module-Short-Description")) {
            return (true);
        }
        if (attrName.equals("OpenIDE-Module-Long-Description")) {
            return (true);
        }
        if (attrName.equals("OpenIDE-Module-Display-Category")) {
            return (true);
        }
        if (attrName.equals("OpenIDE-Module-Provides")) {
            return (true);
        }
        if (attrName.equals("OpenIDE-Module-Requires")) {
            return (true);
        }
        if (attrName.equals("OpenIDE-Module-Recommends")) {
            return (true);
        }
        if (attrName.equals("OpenIDE-Module-Needs")) {
            return (true);
        }
        if (attrName.equals("AutoUpdate-Show-In-Client")) {
            return (true);
        }
        if (attrName.equals("AutoUpdate-Essential-Module")) {
            return (true);
        }

        return (false);

    }

    /**
     * Scans element l10n
     */
    private void scanElementL10N(Element element) { // <l10n>
        l10nProperties = new Properties();

        NamedNodeMap attrs = element.getAttributes();
        for (int i = 0; i < attrs.getLength(); i++) {
            Attr attr = (Attr) attrs.item(i);
            l10nProperties.setProperty(attr.getName(), attr.getValue());
        }
    }

    /**
     * Scans license element
     */
    private void scanElementLicense(Element element) { // <license>
        String licenseName = null;

        NamedNodeMap attrs = element.getAttributes();
        for (int i = 0; i < attrs.getLength(); i++) {
            Attr attr = (Attr) attrs.item(i);
            if (attr.getName().equals("name")) {
                licenseName = attr.getValue();
            }
            moduleProperties.setProperty("license", licenseName);
        }

        NodeList nodes = element.getChildNodes();

        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            switch (node.getNodeType()) {
                case Node.CDATA_SECTION_NODE:
                    Licenses.setLicense(licenseName, ((CDATASection) node).getData(), getName(), getSpecVersion());
                    break;
                case Node.TEXT_NODE:
                    Licenses.setLicense(licenseName, ((Text) node).getData(), getName(), getSpecVersion());
                    break;
            }
        }
    }

    /**
     * Scans description element
     */
    private void scanElementDescription(Element element) { // <description>
        NodeList nodes = element.getChildNodes();

        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            switch (node.getNodeType()) {
                case Node.TEXT_NODE:
                    description = ((Text) node).getData();
                    break;
                case Node.CDATA_SECTION_NODE:
                    description = ((CDATASection) node).getData();
                    break;
            }
        }

    }

    /**
     * Scans notification element
     */
    private void scanElementNotification(Element element) { // <module_notification>
        NodeList nodes = element.getChildNodes();

        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            switch (node.getNodeType()) {
                case Node.TEXT_NODE:
                    notification = ((Text) node).getData();
                    break;
                case Node.CDATA_SECTION_NODE:
                    notification = ((CDATASection) node).getData();
                    break;
            }
        }

    }

    /**
     * Parses dependency string and loads list of dependecies
     */
    private void readDependencies(String depString, String depType) {

        char[] deps = depString.toCharArray();
        int flag = 0;
        int save_flag = 0;
        StringBuffer one_dependency = new StringBuffer();
        Dependency new_dep;

        int n = depString.length();
        for (int i = 0; i < n; i++) {
            if (deps[i] == ' ') {
                continue;
            }

            if (deps[i] == '[' && flag == 0) {
                continue;
            }

            if (deps[i] != '[' && deps[i] != ',' && flag == 0) {
                flag = 1;
            }

            if (deps[i] == ',' && flag == 1) {
                flag = 0;
                save_flag = 1;
            }

            if (deps[i] == ']' && flag == 1) {
                flag = 0;
                save_flag = 1;
            }

            if (deps[i] == '[' && flag == 1) {
                flag = 2;
                save_flag = 1;
            }

            if (flag == 1) {
                one_dependency.append(deps[i]);
            }
            if (save_flag == 1 || (i == n - 1 && flag == 1)) {
//System.out.println("->" + one_dependency.toString() );

                new_dep = new Dependency(one_dependency.toString(), depType, language);
                dependencies.add(new_dep);
                if (depType.compareTo("module_dependency") == 0) {
                    moduleDeps.add(new_dep.module);
                    if (new_dep.version != null) {
                        versionDeps.add(new_dep.module + " " + new_dep.version);
                    /* old version of dependencies storing                   
                    module_dependencies.add(new_dep.module);
                    if ( new_dep.version!=null ) 
                    module_dependencies.add(new_dep.module + " " + new_dep.version);
                     */
                    }
                }

                one_dependency = new StringBuffer();
                save_flag = 0;
            }
        }

    }

    /**
     * Sets params according to External Parameters setup file
     */
    private void applyModulesSetup() {
        String one_param, old_param;
        String name = this.getName();
        String spec_version = this.getSpecVersion();
        String distribution;

// setup of language
        if (language.compareTo("en") != 0) {
            old_param = manifestProperties.getProperty("OpenIDE-Module");
            manifestProperties.setProperty("OpenIDE-Module", old_param + "_" + language);
        }

// setup of os
        one_param = ModulesSetup.getParam(name, spec_version, OS_NAME);
        if (one_param != null) {
            if (language.compareTo("en") == 0) {
                one_param = language + "_" + one_param;
            }
            old_param = manifestProperties.getProperty("OpenIDE-Module");
            manifestProperties.setProperty("OpenIDE-Module", old_param + "_" + one_param);
        }

        name = this.getName();

// setup of distribution url        
        distribution = ModulesSetup.getParam(name, spec_version, DISTRIBUTION_URL);

        if (distribution == null) {
            distribution = createDistribURL();

            if (distribution == null) {
//                distribution = AppSetup.global_property(DISTRIBUTION_BASE);
                one_param = ModulesSetup.getParam(name, spec_version, DISTRIBUTION_DIR);

                if (one_param == null && parentGroup != null) {
                    one_param = parentGroup.distribution_dir;
                }
                if (one_param != null) {
                    distribution += one_param;
                }
                distribution += nbmName;
//                distribution += AppSetup.global_property(DISTRIBUTION_SUFFIX);
            } else {
                distribution += nbmName;
//                distribution += AppSetup.global_property(DISTRIBUTION_SUFFIX);
            }
        }

        moduleProperties.setProperty("distribution", distribution);

// setup of license
        one_param = ModulesSetup.getParam(name, spec_version, LICENSE);
        if (one_param == null && parentGroup != null) {
            one_param = parentGroup.license;
        }
        if (one_param != null) {
            moduleProperties.setProperty("license", one_param);// setup of implementation version
        }
        one_param = ModulesSetup.getParam(name, spec_version, IMPL_VERSION);
        if (one_param != null) {
            manifestProperties.setProperty("OpenIDE-Module-Implementation-Version", one_param);
        }

// setup of homepage
        one_param = ModulesSetup.getParam(name, spec_version, HOMEPAGE);
        if (one_param != null) {
            moduleProperties.setProperty("homepage", one_param);
        }

// setup of per module notification
        String notif_file_path = null;
        String one_line;
        StringBuffer sb = new StringBuffer();

        one_param = ModulesSetup.getParam(name, spec_version + os, MODULE_NOTIFICATION);
        if (one_param == null) {
            one_param = ModulesSetup.getParam(name, spec_version, MODULE_NOTIFICATION);
        }
        if (one_param != null) {
            notif_file_path = UpdateCenter.getNotifPath() + one_param;
        }
        if (notif_file_path != null) {
            try {
                BufferedReader in = new BufferedReader(new FileReader(notif_file_path));
                while ((one_line = in.readLine()) != null) {
                    sb.append(one_line);
                }
                this.notification = sb.toString();

                in.close();
            } catch (FileNotFoundException e) {
                System.out.println("Failure. Can't find notification " + notif_file_path);
            } catch (IOException e) {
                System.out.println("Failure. Can't read notification " + notif_file_path);
            }
        }

// setup of additional dependencies
        String add_deps = ModulesSetup.getParam(name, spec_version, ADD_DEPS);

        if (add_deps != null) {
            String org_deps = manifestProperties.getProperty("OpenIDE-Module-Module-Dependencies");
            StringBuffer all_deps;

            if (org_deps != null) {
                all_deps = new StringBuffer(org_deps);
                all_deps.append(", ");
                all_deps.append(add_deps);
            } else {
                all_deps = new StringBuffer(add_deps);
            }

            manifestProperties.setProperty("OpenIDE-Module-Module-Dependencies", all_deps.toString());

            readDependencies(add_deps, "module_dependency");
        }
    }

    /**
     * Constructs distribution URL for Netbeans NBM
     */
    private String createDistribURL() {
        ModuleGroup one_group = this.parentGroup;
        StringBuffer dist_url = new StringBuffer();
        while (one_group != null) {
            dist_url.insert(0, one_group.name + "/");
            one_group = one_group.getParentGroup();
        }
        String base_url = UpdateCenter.getDistributionBaseURL();
        if ((base_url != null) && (!base_url.trim().equals(""))) {
            dist_url.insert(0, base_url + "/");
        }
        return (dist_url.toString());
    }

    private void setConfigData(NBMfile nbmFile) {
        ModuleConfigData config_data = new ModuleConfigData(nbmFile, getCodebaseName());

        String eager_value = config_data.getEagerValue();
        if (eager_value != null) {
            moduleProperties.setProperty("eager", eager_value);
        }
        String autload_value = config_data.getAutoloadValue();
        if (autload_value != null) {
            moduleProperties.setProperty("autoload", autload_value);
        }
    }
}
