/**
 * Generate a sdoc from the Dojo documentation (extracted from the sources)
 */
package dojosdoc;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Extract documentation about Dojo from the sources
 * Look for "summary:" items; the prototype should be above
 *
 * No need to generate compatibility flags here since it seems
 * that it includes everything:
 *   http://dojotoolkit.org/pipermail/dojo-interest/2006-January/003077.html
 *
 * @todo Handle classes: When you see this create a class entry! 
 *    dojo.declare("dojo.data.api.Identity", dojo.data.api.Read, {
 *            //      summary:
 * 
 * @author Tor Norbye
 */
public class Dojo {
    private static final boolean SKIP_PRIVATE = true;

    public static void main(String[] args) {
        if (args.length != 1) {
            System.err.println("Usage: Main <dojo-svn-tree/dojo/trunk>");
        }
        File dojo = new File(args[0]);
        if (!dojo.exists()) {
            System.err.println(dojo.getPath() + " does not exist");
            System.exit(0);
        }


        try {
            List<DojoFile> files = new ArrayList<DojoFile>();
            new Dojo().process(dojo, files, "");

            //File root = new File("build/dojo");
            //root.mkdirs();
            File dest = new File("build", "dojo.sdoc");
            dest = dest.getAbsoluteFile();
            File parent = dest.getParentFile();
            parent.mkdirs();
            Writer writer = new BufferedWriter(new FileWriter(dest));
            writeHeader(writer);

            for (DojoFile f : files) {
//                String name = f.relativeFile;
//                assert name.endsWith(".js");
//                name = name.substring(0, name.length() - 2) + "sdoc";
                //File dest = new File(root, name);
//                dest = dest.getAbsoluteFile();
//                File parent = dest.getParentFile();
//                parent.mkdirs();
//                Writer writer = new BufferedWriter(new FileWriter(dest));
//                writeHeader(writer);

                for (String s : f.comments) {
                    writer.write(s);
                    writer.write("\n");
                }

//                writer.close();
            }
            writer.close();
        } catch (IOException ex) {
            Logger.getLogger(Dojo.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    class DojoFile {

        File file;
        String relativeFile;
        String copyright;
        List<String> comments;
        Map<String, String> nextLines;
    }

    private static void writeHeader(Writer writer) throws IOException {
        writer.write("//\n");
        writer.write("// For full details see http://dojotoolkit.org/about/license\n");
        writer.write("//\n");
        writer.write("// Dojo is available under *either* the terms of the modified BSD license *or* the\n");
        writer.write("// Academic Free License version 2.1. As a recipient of Dojo, you may choose which\n");
        writer.write("// license to receive this code under (except as noted in per-module LICENSE\n");
        writer.write("// files). Some modules may not be the copyright of the Dojo Foundation. These\n");
        writer.write("// modules contain explicit declarations of copyright in both the LICENSE files in\n");
        writer.write("// the directories in which they reside and in the code itself. No external\n");
        writer.write("// contributions are allowed under licenses which are fundamentally incompatible\n");
        writer.write("// with the AFL or BSD licenses that Dojo is distributed under.\n");
        writer.write("//\n\n");

        writer.write("//\n// Doc definitions for Dojo support in NetBeans, \n" +
                "// automatically generated from the Dojo sources\n" +
                "// using the generator in misc/javascript.generatestubs/dojo-sdoc\n" +
                "// Version: 1.0\n" +
                "//\n\n");
    }

    void process(File f, List<DojoFile> files, String relative) throws IOException {
        if (f.isDirectory()) {
            if (f.getName().equals("tests") || f.getName().equals(".svn")) { // NOI18N
                return;
            }

            for (File c : f.listFiles()) {
                String childDir = relative.length() > 0 ? relative + "/" + c.getName() : c.getName();
                process(c, files, childDir);
            }

            return;
        }
        
        if (!f.getName().endsWith(".js")) {
            return;
        }

        if (f.getName().endsWith("-min.js") || f.getName().endsWith("-debug.js")) {
            assert false : f.getPath();
            return;
        }

        currentFile = f;
        currentNamespace = null;
        String js = readFile(f);

        List<String> comments = findComments(js);
        if (comments.size() > 0) {
            DojoFile dojoFile = new DojoFile();
            dojoFile.file = f;
            dojoFile.relativeFile = relative;
            dojoFile.comments = comments;

            files.add(dojoFile);
        }
    }

    enum ElementType {

        FUNCTION, PROPERTY, EVENT, CONSTRUCTOR
    };

    private String fixTypos(String s) {
        // Fix typos I've found in the current YUI release (2.5.0)
        // @propery
        // @depreciated
        // @parem
        // @parm

        if (s.indexOf("@propery") != -1) {
            s = s.replaceAll("@propery", "@property");
        }

        if (s.indexOf("@parem") != -1) {
            s = s.replaceAll("@parem", "@param");
        }

        if (s.indexOf("@parm") != -1) {
            s = s.replaceAll("@parm", "@param");
        }

        if (s.indexOf("@depreciated") != -1) {
            s = s.replaceAll("@depreciated", "@deprecated");
        }

        // Normalize case:
        // @Class, @Extends, @TODO
        if (s.indexOf("@Class") != -1) {
            s = s.replaceAll("@Class", "@class");
        }
        if (s.indexOf("@Extends") != -1) {
            s = s.replaceAll("@Extends", "@extends");
        }
        if (s.indexOf("@TODO") != -1) {
            s = s.replaceAll("@TODO", "@todo");
        }

        return s;
    }

    private String removeLinks(String s) {
        // Remove <a ...></a> stuff
        int lt = -1;
        int lta = s.indexOf("<a");
        int ltb = s.indexOf("</a");
        if (lta == -1) {
            if (ltb == -1) {
                return s;
            } else {
                lt = ltb;
            }
        } else {
            if (ltb == -1) {
                lt = lta;
            } else {
                lt = Math.min(lta, ltb);
            }
        }
        if (lt == -1) {
            return s;
        }

        int lte = s.indexOf('>', lt);
        if (lte == -1) {
            return s;
        }
        assert s.indexOf('<', lt + 1) == -1 || s.indexOf('<', lt + 1) > lte;

        s = s.substring(0, lt) + s.substring(lte + 1);
        return removeLinks(s);
    }

    private String fixTypes(String s) {
        // Change type declarations from Int etc. to Number and so on


        // Some errors in the spec
        if (s.indexOf("|HTML Element}") != -1) {
            s = s.replace("|HTML Element}", "|HTMLElement}");
        }
        if (s.indexOf("{String or HTMLEle") != -1) {
            s = s.replace("{String or HTMLEle", "{String|HTMLEle");
        }
        if (s.indexOf("[int, int, int]") != -1) {
            s = s.replace("[int, int, int]", "Number[]");
        }
        if (s.indexOf("[int, int]") != -1) {
            s = s.replace("[int, int]", "Number[]");
        }

        int index = 0;
        docParam:
        while (index < s.length()) {
            // Find next type string
            int nextParam = s.indexOf("@param", index);
            int nextType = s.indexOf("@type", index);
            int next;
            if (nextParam == -1) {
                next = nextType;
            } else if (nextType == -1) {
                next = nextParam;
            } else {
                next = Math.min(nextParam, nextType);
            }
            if (next == -1) {
                break;
            }

            boolean isParam = next == nextParam;
            index = next + (isParam ? "@param".length() : "@type".length());

            // Iterate over all the @type params in the comma list
            while (index < s.length()) {

                for (; index < s.length(); index++) {
                    char c = s.charAt(index);
                    if (c == '\n') {
                        continue docParam;
                    }
                    if (!Character.isWhitespace(c)) {
                        break;
                    }
                }

                if (isParam) {
                    // Find the opening '{'
                    if (s.charAt(index) != '{') {
                        // Nope, this @param doesn't specify the type
                        continue docParam;
                    }
                }

                int begin = index;

                // Find the end
                for (; index < s.length(); index++) {
                    char c = s.charAt(index);
                    if (c == '}') {
                        break;
                    }
                    if (c == '\n') {
                        continue docParam;
                    }
                    if (Character.isWhitespace(c)) {
                        break;
                    }
                }
                int end = index;

                String type = s.substring(begin, end);
                String translatedType = translateType(type);
                if (translatedType != null) {
                    s = s.substring(0, begin) + translatedType + s.substring(end);
                    // Recurse
                    return fixTypes(s);
                }

                if (!isParam) {
                    break;
                }

                if (s.charAt(index) == '}') {
                    break;
                }

                // Skip spaces
                for (; index < s.length(); index++) {
                    char c = s.charAt(index);
                    if (!Character.isWhitespace(c)) {
                        break;
                    }
                }
                if (s.charAt(index) == '}') {
                    break;
                }
                if (s.charAt(index) == '/' || s.charAt(index) == ',' || s.charAt(index) == '|') {
                    index++;
                } else {
                    System.err.println("WARNING: Unexpected page - " + s.charAt(index) + " in " + s.substring(index - 10, index + 10));
                    //assert false : s.charAt(index) + " in " + s.substring(index-10, index+10);
                    return s;
                }

            }
        }

        return s;
    }

    private String translateType(String s) {
        // object => Object
        // Int => Number
        // int => Number
        // string => String
        // function => Function
        // boolean => Boolean
        // float => Float

        // Not yet done:
        // Variant => ?
        // String[] => Array<String>

        // Not sure what to do about this:
        // @type string|{searchExp: string, replaceStr: string}
        // @type [int, int]
        // @type {"x": int, "y": int}

        if ("object".equals(s)) {
            return "Object";
        }
        if ("Int".equals(s)) {
            return "Number";
        }
        if ("int".equals(s)) {
            return "Number";
        }
        if ("string".equals(s)) {
            return "String";
        }
        if ("function".equals(s)) {
            return "Function";
        }
        if ("boolean".equals(s)) {
            return "Boolean";
        }
        if ("float".equals(s)) {
            return "Number";
        }

        return null;
    }

    private List<String> findComments(String js) {
        List<String> comments = new ArrayList<String>();
        
        int namespaceDecl = js.indexOf("dojo.provide(\"");
        if (namespaceDecl != -1) {
            int namespaceEnd = js.indexOf('"', namespaceDecl+14);
            assert namespaceEnd != -1;
            currentNamespace = js.substring(namespaceDecl+14, namespaceEnd);
            assert currentNamespace.indexOf(" ") == -1;
        }
        
        if (js.indexOf("summary:") == -1) {
            if (js.indexOf("summary") != -1) {
                // There are some typos in the documentation
                // where it's missing the :
                //./_base/fx.js:34:		//	summary
                //./_base/html.js:841:		// summary
                //./_firebug/firebug.js:306:		// summary
                //./fx.js:304:	// summary
                //./fx.js:345:	// summary
                //./fx.js:372:	// summary
                //./rpc/RpcService.js:68:		// summary
                
                // TODO - handle this somehow
                String old = js;
                js = js.replace("// summary\n", "// summary:\n");
                if (!js.equals(old)) {
                    return findComments(js);
                }
            }
            return comments;
        }

        int index = 0;
        while (index < js.length()) {
            int next = js.indexOf("summary:", index);
            if (next == -1) {
                return comments;
            }

            // The line above should be the signature

            // The next lines should constitute the comment block
            String comment = extractComment(js, next);
            if (comment != null) {
                comments.add(comment);
            }

            index = next + 5;
        }
        
        return comments;
    }
    
    private File currentFile;
    private String currentNamespace;
    
    final static Pattern KEY_PATTERN = Pattern.compile("//\\s*(\\S+)\\:");

    private String extractComment(String js, int summaryIndex) {
        ItemType itemType = null;
        
        int prototypeBegin = -1;
        int prototypeEnd = js.lastIndexOf("\n", summaryIndex);
        if (prototypeEnd != -1) {
            prototypeBegin = js.lastIndexOf("\n", prototypeEnd - 1);
        }
        String prototypeLine = "";
        if (prototypeBegin != -1) {
            assert prototypeBegin != -1 : " offset " + prototypeEnd + " in " + currentFile.getPath();
            prototypeLine = js.substring(prototypeBegin + 1, prototypeEnd);
            prototypeLine = prototypeLine.trim();
        }

        if (currentFile.getName().equals("dojo.js") && summaryIndex == 3) {
            prototypeLine = "dojo = function() {}";
            itemType = ItemType.PROPERTY;
        }
        
        // Known exceptions
        if (currentFile.getName().equals("bootstrap.js")) {
            if (prototypeLine.equals("djConfig = {") || prototypeLine.indexOf("dojo.locale = {") != -1) {                    
                return null;
            }
        }

        if (currentFile.getName().equals("robot.js")) {
            if (prototypeLine.indexOf("scrollIntoView : function") != -1 || prototypeLine.indexOf("mouseMoveAt : function") != -1) {                    
                return null;
            }
        }

        if (currentFile.getName().equals("OpenAjax.js")) {
            // No documentation here anyway
            return null;
        }
        if (currentFile.getName().equals("hostenv_rhino.js")) {
            if (prototypeLine.length() == 0) {
                return null;
            }
            if (prototypeLine.indexOf("function setTimeout(func, delay){") != -1) {
                return null;
            }
        }
        if (currentFile.getName().equals("Manager.js") && prototypeLine.length() == 0) {
            return null;
        }
        if ("// FIREBUG LITE".equals(prototypeLine)) {
            return null;
        }
        
        assert prototypeLine.length() > 0 : " empty prototype around index " + summaryIndex + " in " + currentFile.getPath() + "  ..." + js.substring(summaryIndex-30, Math.min(js.length(),summaryIndex+60)) + "...";

        if (prototypeLine.startsWith("d.")) {
            // Alias for dojo in many construction methods
            prototypeLine = "dojo." + prototypeLine.substring(2);
        }

        if (!(prototypeLine.startsWith("dojo.") || prototypeLine.startsWith("dojo[\""))) {

            if (prototypeLine.startsWith("_")) {
                return null;
            }
            int privateIndex = prototypeLine.indexOf("._");
            if (privateIndex != -1 && privateIndex < prototypeLine.indexOf("=")) {
                // Don't worry about private stuff
                return null;
            }
            
            // Look backwards -- perhaps it's a line that should be combined... like this one in array.js:
            //indexOf: function(      /*Array*/               array,
            //                                        /*Object*/              value,
            //                                        /*Integer?*/    fromIndex,
            //                                        /*Boolean?*/    findLast){
            if (prototypeLine.indexOf("function") == -1) {
                int funcStart = js.lastIndexOf("function", prototypeBegin);
                assert funcStart > prototypeBegin-200 : " in " + prototypeLine + " from " + currentFile.getPath();
                int newLineBegin = js.lastIndexOf('\n', funcStart)+1;
                String lines = js.substring(newLineBegin, prototypeEnd).trim();
                lines = lines.replace('\n', ' ');
                if (lines.matches("\\S+\\:.*$")) {
                    String ns = currentNamespace;
                    if (ns == null) {
                        ns = "dojo";
                    }
                    lines = ns + "." + lines.substring(0, lines.indexOf(':')) + " = " + lines.substring(lines.indexOf("function"));
                }
                prototypeLine = lines;
            } else if (prototypeLine.matches("\\S+\\:\\s*function.*$")) {
                    String ns = currentNamespace;
                    if (ns == null) {
                        ns = "dojo";
                    }
                prototypeLine = ns + "." + prototypeLine.substring(0, prototypeLine.indexOf(':')) + " = " + prototypeLine.substring(prototypeLine.indexOf("function"));
            }
            
        } else if (prototypeLine.startsWith("dojo.declare(\"")) {
            // TODO - handle these clases
            return null;
        }
        
        if (prototypeLine.startsWith("this.")) {
            return null;
        } else if (prototypeLine.startsWith("var ")) {
            return null;
        } else if (prototypeLine.startsWith("function") && !prototypeLine.startsWith("function dojo")) {
            return null;
        }
        if (prototypeLine.startsWith("_")) {
            return null;
        }
        
        if (prototypeLine.startsWith("test.") || prototypeLine.startsWith("tests.")) {
            return null;
        }
        
        if (currentFile.getName().equals("back.js") && prototypeLine.startsWith("back.")) {
            prototypeLine = "dojo." + prototypeLine;
        }
        
        assert prototypeLine.startsWith("dojo =") || prototypeLine.startsWith("dojo.") || prototypeLine.startsWith("dojo[\"") : "\n\n" + prototypeLine + " in " + currentFile.getPath();
        
        // See if it's one of the special sections:
        // /*=====
        int prevLine = js.lastIndexOf("\n", prototypeBegin-1);
        if (prevLine == -1) {
            prevLine = 0;
        } else {
            prevLine++;
        }
        if (js.startsWith("/*=====", prevLine)) {
            return null;
        }

        List<String> keys = new ArrayList<String>();
        Map<String, String> values = new HashMap<String, String>();
        int lineBegin = prototypeEnd + 1;
        String currentKey = null;
        StringBuilder currentValue = new StringBuilder();
        int lineEnd = js.indexOf('\n', lineBegin);
        for (; lineBegin < js.length(); lineBegin = lineEnd+1) {
            lineEnd = js.indexOf('\n', lineBegin);
            assert lineEnd != -1;
            String line = js.substring(lineBegin, lineEnd).trim();
            if (!line.startsWith("//")) {
                break;
            }
            Matcher m = KEY_PATTERN.matcher(line);

            if (m.find()) {
                if (currentValue.length() > 0) {
                    assert currentKey != null : currentValue + " in " + currentFile.getPath();
                    values.put(currentKey, currentValue.toString());
                    currentValue.setLength(0);
                }
                currentKey = m.group(1);
                keys.add(currentKey);
                int colon = line.indexOf(':');
                assert colon != -1 : line;
                String rest = line.substring(colon+1).trim();
                if (rest.length() > 0) {
                    currentValue.append(rest);
                    currentValue.append("\n");
                }
            } else {
                currentValue.append(line.substring(2).trim()); // 2: skip //

                currentValue.append("\n"); // Preserve individual lines (for the code example sections)

            }
        }
        if (currentValue.length() > 0) {
            assert currentKey != null : currentValue + " in " + currentFile.getPath();
            values.put(currentKey, currentValue.toString());
        }

        Function function = analyzePrototype(prototypeLine);
        assert function != null : prototypeLine;
        
        String id = function.id;
        String name = function.name;
        
        if (SKIP_PRIVATE && name.startsWith("_")) {
            // Private ... just skip
            return null;
        }
        List<String> parameters = function.parameters;
        Map<String, String> types = function.types;

        StringBuilder sb = new StringBuilder();
        Set<String> handled = new HashSet<String>();

        sb.append("/**\n");
        String summary = values.get("summary");
        if (summary != null) {
            handled.add("summary");
            if (Character.isLetter(summary.charAt(summary.length()-1)) || summary.endsWith("\"")) {
                summary = summary+".";
            }
            appendFormatted(sb, summary, false);
            sb.append(" *\n");
        }
        String desc = values.get("description");
        if (desc != null) {
            handled.add("description");
            if (Character.isLetter(desc.charAt(desc.length()-1)) || desc.endsWith("\"")) {
                desc = desc+".";
            }
            appendFormatted(sb, desc, false);
        }
        
        sb.append(" * Defined in file " + currentFile.getName() + "\n *\n");
        
        String ret = values.get("return");
        if (ret != null) {
            handled.add("return");
            sb.append(" * @return ");
            if (ret.startsWith("a String")) {
                sb.append("{String} ");
            } else if (ret.startsWith("dojo.NodeList") || ret.startsWith("`dojo.NodeList`")) {
                sb.append("{dojo.NodeList} ");
            } else if (ret.indexOf("return `true`") != -1){
                sb.append("{Boolean} ");
            } else if (ret.indexOf("an array of strings") != -1) {
                sb.append("{String[]} ");
            }
            appendFormatted(sb, ret, true);
        }
        String code = values.get("example");
        if (code != null) {
            handled.add("example");
            sb.append(" * @example\n");
            for (String s : code.split("\n")) {
                sb.append(" * ");
                if (s.startsWith("|")) {
                    assert s.startsWith("|") : code + " from " + currentFile.getPath();
                    sb.append(s.substring(1));
                } else {
                    sb.append(s);
                }
                sb.append("\n");
            }
        }

        sb.append(" * @id ");
        sb.append(id);
        sb.append("\n");

        if (itemType != null && itemType != function.itemType) {
            function.itemType = itemType;
        }
        if (function.itemType == ItemType.FUNCTION) {
            sb.append(" * @method " + name + "\n");
        } else if (function.itemType == ItemType.PROPERTY) {
            sb.append(" * @property " + name + "\n");
        } else {
            assert false : function;
        }
        
        if (function.name.startsWith("_")) {
            sb.append(" * @private\n");
        }

        // Create parameter documentation
        for (String parameter : parameters) {
            sb.append(" * @param ");
            String type = types.get(parameter);
            if (type != null) {
                sb.append("{");
                sb.append(type);
                sb.append("} ");
            }
            sb.append(parameter);
            sb.append(" ");
            String paramDesc = values.get(parameter);
            handled.add(parameter);
            if (paramDesc != null) {
                appendFormatted(sb, paramDesc, true);
            } else {
                sb.append("\n");
            }
        }

        // TODO - return values? Their types apparently aren't there...

        for (String key : keys) {
            if (handled.contains(key)) {
                continue;
            }

//            assert false : key + " : " + values.get(key) + " : " + function + " from " + currentFile.getPath();
            System.err.println("WARNING: Encountered an unexpected key: " + key + " : " + values.get(key) + " : " + function + " from " + currentFile.getPath());
        }

        if (sb.charAt(sb.length()-1) != '\n') {
            sb.append("\n");
        }
        sb.append("*/\n");
        return sb.toString();
    }
    
    private Function analyzePrototype(String line) {
        line = line.trim();
        if (line.startsWith("dojo[\"")) {
            line = "dojo." + line.substring(6, line.indexOf('\"', 6)) + line.substring(line.indexOf('\"', 6)+2);
        }
        assert line.startsWith("dojo");
        
        Function f = new Function(line);
        
        int rhs = line.indexOf('=');
        assert rhs != -1 : rhs + " no rhs in " + line + " from " + currentFile.getPath();
        int end = rhs-1;
        for (; end >= 0; end--) {
            if (!Character.isWhitespace(line.charAt(end))) {
                break;
            }
        }
        assert end >= 0;

        String id = line.substring(0, end+1);
        f.id = id;
        int lastDot = id.lastIndexOf('.');
        if (lastDot != -1) {
            f.name = id.substring(lastDot+1);
        } else {
            f.name = id;
        }
        
        if (line.indexOf("function", rhs) != -1) {
            f.itemType = ItemType.FUNCTION;
            
            // Find parameters
            int paramIndex = line.indexOf("function", rhs);
            assert paramIndex != -1;
            paramIndex = line.indexOf("(", paramIndex);
            assert paramIndex != -1 : line;
            int endIndex = line.indexOf(")", paramIndex);
            assert endIndex != -1 : line;
            String paramString = line.substring(paramIndex+1, endIndex);
            for (String param : paramString.split(",")) {
                param = param.trim();
                String name;
                if (param.startsWith("/*") && !param.startsWith("/*====")) {
                    int endType = param.indexOf("*/");
                    assert endType != -1 : param;
                    name = param.substring(endType+2).trim();
                    String type = param.substring(2, endType).trim();
                    f.types.put(name, type);
                } else {
                    name = param;
                }
                f.parameters.add(name);
            }
        } else {
System.err.println("WARNING: Is this a property? " + line);
            f.itemType = ItemType.PROPERTY;
        }
        
        
        return f;
    }

    enum ItemType { FUNCTION, PROPERTY };
    
    class Function {
        Function(String prototype) {
            this.prototype = prototype;
        }
        private String prototype;
        private String id;
        private String name;
        private ItemType itemType;
        private List<String> parameters = new ArrayList<String>();
        private Map<String,String> types = new HashMap<String, String>();
        
        @Override
        public String toString() {
            return "Function(" + prototype + "; id=" + id + ")";
        }
    }

    private void appendFormatted(StringBuilder sb, String v, boolean skipFirstPrefix) {
        for (String s : v.split("\n")) {
            if (!skipFirstPrefix) {
                sb.append(" * ");
            }
            skipFirstPrefix = false;
            sb.append(s);
            sb.append("\n");
        }
    }

    private String readFile(File file) {
        StringBuilder sb = new StringBuilder();
        try {
            BufferedReader f = new BufferedReader(new FileReader(file));
            while (true) {
                String line = f.readLine();
                if (line == null) {
                    break;
                }
                sb.append(line);
                sb.append("\n");
            }
            f.close();
        } catch (IOException e) {
            e.printStackTrace();
            System.exit(1);
        }

        return sb.toString();
    }
}

