
package org.netbeans.jsstubsgenerator.stub;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.netbeans.jsstubsgenerator.core.mdn.ReturnType;
import org.netbeans.jsstubsgenerator.print.JsConstruct;
import org.netbeans.jsstubsgenerator.print.JsFunction;
import org.netbeans.jsstubsgenerator.print.JsObject;
import org.netbeans.jsstubsgenerator.print.JsProperty;

/**
 * Represents one stub item (method, object, parameter, const).
 *
 * @author Martin Fousek <marfous@netbeans.org>
 */
public class StubObject implements Cloneable {

    public static final String GLOBAL = "global";


    // cannot be final since it's required to change that for interfaces implementation
    private Origin origin;
    private String parent;
    private String name;
    private Type type;
    private String summary = "";
    private List<String> since = new LinkedList<String>();
    private String syntax = "";
    private String description = "";
    private String superClass;
    private Set<String> implementIfces = new HashSet<String>();
    private boolean instance;
    private Return returns;
    private boolean isStatic;
    private boolean isDeprecated;
    private List<Parameter> parameters = new ArrayList<Parameter>();
    private Map<String, StubObject> attributes = new HashMap<String, StubObject>();

    // says whether the object must be pruned for bundled documentation
    private boolean prune = false;
    // whether can be parsed enough information for this stub object or not
    private boolean uncompleted = false;
    // whether it's underscored attr. - means commonly not standard stub
    private boolean underscored = false;
    
    private boolean nonstandard = false;

    private static final Logger LOGGER = Logger.getLogger(StubObject.class.getName());

    public StubObject(Origin origin, String name, String parent) {
        this.origin = origin;
        this.parent = parent;
        this.name = name;
    }
    
    public void setNonStandard(boolean isnonstandard) {
         nonstandard = isnonstandard;
    }
    
    public boolean isNonStandard() {
         return nonstandard;
    }
    
    public void setDeprecated(boolean isDeprecated) {
        this.isDeprecated = isDeprecated;
    }
    
    public boolean isDeprecated() {
        return isDeprecated;
    }

    public boolean isPrune() {
        return prune;
    }

    public void setPrune(boolean prune) {
        this.prune = prune;
    }

    public void setOrigin(Origin origin) {
        this.origin = origin;
    }

    public void setDescription(String description) {
        this.description = description;
    }

    public void setParent(String parent) {
        this.parent = parent;
    }

    public void addSince(String since) {
        this.since.add(since);
    }

    public void addParameter(Parameter parameter) {
        parameters.add(parameter);
    }

    public String getSummary() {
        return summary;
    }

    public String getParent() {
        return parent;
    }

    public Return getReturns() {
        return returns;
    }

    public Set<String> getImplements() {
        return implementIfces;
    }

    public void setSuperClass(String superClass) {
        this.superClass = superClass;
    }

    public boolean isInstance() {
        return instance;
    }

    public void setInstance(boolean instance) {
        this.instance = instance;
    }

    public String getSuperClass() {
        return superClass;
    }

    public boolean isStatic() {
        return isStatic;
    }

    public void setStatic(boolean isStatic) {
        this.isStatic = isStatic;
    }

    public void setReturns(Return returns) {
        this.returns = returns;
    }
    
    public void setSummary(String summary) {
        this.summary = summary;
    }

    public void setSyntax(String syntax) {
        this.syntax = syntax;
    }

    public void putImplements(String inteface) {
        implementIfces.add(inteface);
    }

    public void putAttribute(StubObject attr) {
        attributes.put(attr.getName(), attr);
    }

    public void putProperties(List<String> properties) {
        for (String property : properties) {
            if (!property.contains("__")) {
                String propertyName = property;
                if (property.lastIndexOf(".") > 0) {
                    propertyName = property.substring(property.lastIndexOf(".")+1);
                }
                StubObject attrStub = new StubObject(origin, propertyName, name);
                if (property.contains("prototype")) {
                    attrStub.setStatic(false);
                } else {
                    attrStub.setStatic(true);
                }
                attrStub.setType(Type.PROPERTY);
                attributes.put(property, attrStub);
            }
        }
    }

    public void putMethods(List<String> methods) {
        for (String method : methods) {
            if (!method.contains("__")) {
                String methodName = method;
                if (method.lastIndexOf(".") > 0) {
                    methodName = method.substring(method.lastIndexOf(".")+1);
                    methodName = methodName.replace("(", "");
                    methodName = methodName.replace(")", "");
                }
                StubObject attrStub = new StubObject(origin, methodName, name);
                if (method.contains("prototype")) {
                    attrStub.setStatic(false);
                } else {
                    attrStub.setStatic(true);
                }
           
                attrStub.setType(Type.METHOD);
                attributes.put(method, attrStub);
            }
        }
    }

    public void putInstanceProperties(List<String> properties) {
        for (String property : properties) {
            if (!property.contains("__")) {
                StubObject attrStub = new StubObject(origin, property, name);
                attrStub.setType(Type.PROPERTY);
                attributes.put(property, attrStub);
            }
        }
    }

    public void putInstanceMethods(List<String> methods) {
        for (String method : methods) {
            if (!method.contains("__")) {
                StubObject attrStub = new StubObject(origin, method, name);
                attrStub.setType(Type.METHOD);
                attributes.put(method, attrStub);
            }
        }
    }

    public void assignAttribute(String attrName, StubObject attrObject) {
        boolean assigned = false;
        for (Entry<String, StubObject> entry : attributes.entrySet()) {
            if (entry.getKey().equalsIgnoreCase(attrName)
                    || entry.getKey().equalsIgnoreCase("__" + attrName + "__")) {
                entry.getValue().merge(attrObject);
                assigned = true;
            }
        }
        // log just for valid, completed, not-just-FF-compatible (not underscored) attrObject
        if (!assigned && !attrObject.isUncompleted()) {
            LOGGER.log(Level.SEVERE, "Attribute where to assign was not found stubName={0}, attrName={1}", new Object[]{name, attrName});
        }
    }

    public StubObject getAttribute(String attrName) {
        return attributes.get(attrName);
    }

    public Map<String, StubObject> getAttributes() {
        return attributes;
    }

    public void setAttributes(Map<String, StubObject> attributes) {
        this.attributes = attributes;
    }

    public void setUncompleted(boolean uncompleted) {
        this.uncompleted = uncompleted;
    }

    public boolean isUncompleted() {
        return uncompleted;
    }

    public void setUnderscored(boolean underscored) {
        this.underscored = underscored;
    }

    public boolean isUnderscored() {
        return underscored;
    }

    public void setType(Type type) {
        this.type = type;
    }

    public void setName(String name) {
        this.name = name;
    }

    public Type getType() {
        return type;
    }

    public Origin getOrigin() {
        return origin;
    }

    public String getName() {
        return name;
    }

    public List<String> getSince() {
        return since;
    }

    public String getSyntax() {
        return syntax;
    }

    public String getDescription() {
        return description;
    }

    public List<Parameter> getParameters() {
        return parameters;
    }

    @Override
    public String toString() {
        return "name=" + name;
    }

    public String dump() {
        StringBuilder sb = new StringBuilder();
        sb.append("===============================================\n");
        sb.append(getStubFileName(origin, name)).append("\n");
        sb.append("summary=").append(summary).append("\n");
//        sb.append("syntax=").append(syntax).append("\n");
        sb.append("description=").append(description).append("\n");
        sb.append("=parameters=").append("\n").append(dumpParams(parameters));
        if (type == Type.OBJECT) {
            sb.append(dumpAttributes());
        }
        return sb.toString();
    }

    private static String getStubFileName(Origin origin, String name) {
        return origin + "_" + name + ".js";
    }

    private String dumpParams(List<Parameter> list) {
        StringBuilder sb = new StringBuilder();
        for (Parameter entry : list) {
            sb.append(entry.getName()).append(" : ");
            if (entry.getType() != null) {
                sb.append(entry.getType().toString());
            }
            sb.append("\n");
        }
        return sb.toString();
    }

    private String dumpAttributes() {
        StringBuilder sb = new StringBuilder();
        sb.append("=properties=").append("\n");
        sb.append(dumpAttribute(true, Type.PROPERTY));
        sb.append("=methods=").append("\n");
        sb.append(dumpAttribute(true, Type.METHOD));
        sb.append("=instance properties=").append("\n");
        sb.append(dumpAttribute(false, Type.PROPERTY));
        sb.append("=instance methods=").append("\n");
        sb.append(dumpAttribute(false, Type.METHOD));
        return sb.toString();
    }

    private String dumpAttribute(boolean isStatic, Type type) {
        StringBuilder sb = new StringBuilder();
        for (Map.Entry<String, StubObject> entry : attributes.entrySet()) {
            StubObject attribute = entry.getValue();
            if (attribute.isStatic() == isStatic && isStatic == true 
                    && type == attribute.getType() && type == Type.PROPERTY) {
                sb.append(dumpEntry(entry));
            }
            if (attribute.isStatic() == isStatic && isStatic == true
                    && type == attribute.getType() && type == Type.METHOD) {
                sb.append(dumpEntry(entry));
            }
            if (attribute.isStatic() == isStatic && isStatic == false
                    && type == attribute.getType() && type == Type.PROPERTY) {
                sb.append(dumpEntry(entry));
            }
            if (attribute.isStatic() == isStatic && isStatic == false
                    && type == attribute.getType() && type == Type.METHOD) {
                sb.append(dumpEntry(entry));
            }
        }
        return sb.toString();
    }

    private String dumpEntry(Entry<String, StubObject> entry) {
        StringBuilder sb = new StringBuilder();
        if (entry.getValue() != null && entry.getValue() != null) {
            sb.append(entry.getValue().toString());
            sb.append("\n");
        } else {
            LOGGER.log(Level.WARNING, "Attribute not loaded: stubName={0}, attrName={1}", new Object[]{name, entry.getKey()});
        }
        return sb.toString();
    }

    public String toDoc(boolean innerAttribute, String superClass) {
        JsConstruct construct;

        if (type == Type.METHOD) {
            construct = new JsFunction(this, 0, innerAttribute);
        } else if (type == Type.PROPERTY) {
            construct = new JsProperty(this, 0, innerAttribute, superClass);
        } else {
            construct = new JsObject(this, 0);
        }
        return construct.print();
    }

    public String toDoc(boolean innerAttribute) {
        return toDoc(innerAttribute, null);
    }

    public String toReturnHolder() {
        StringBuilder sb = new StringBuilder();
        if (!attributes.isEmpty()) {
            System.out.println("ReturnHolder[] " + name + "Holders = new ReturnHolder[] {");
            for (Entry<String, StubObject> entry : attributes.entrySet()) {
                StubObject attr = entry.getValue();
                System.out.println("new ReturnHolder(\"" + attr.getName() + "\", \"" + ReturnType.guessReturns(getName(), attr.getName(), attr.getSummary()) + "\"),");
            }
            System.out.println("};");
            System.out.println("RETURN_TYPES.put(\""+ name +"\", Arrays.asList(" + name + "Holders));");
        }
        return sb.toString();
    }

    private void merge(StubObject attrObject) {
        summary = attrObject.getSummary();
        if (!since.equals(attrObject.getSince())) 
            since = attrObject.getSince();
        syntax = attrObject.getSyntax();
        description = attrObject.getDescription();
        returns = attrObject.getReturns();
        parameters = attrObject.getParameters();
        attributes = attrObject.getAttributes();
    }

    /**
     * XXX - cloning is used just for changing the object parent, no all properties are cloned correctly - like
     * parameters, attributes.
     */
    @Override
    public StubObject clone() throws CloneNotSupportedException {
        StubObject clone = new StubObject(origin, name, parent);
        clone.type = type.valueOf(type.name());
        clone.summary = summary;
        clone.since = new LinkedList<String>(since);
        clone.syntax = syntax;
        clone.description = description;
        clone.superClass = superClass;
        clone.implementIfces = new HashSet<String>(implementIfces);
        clone.instance = instance;
        clone.returns = new Return(returns.getType(), returns.getDesc());
        clone.isStatic = isStatic;
        clone.parameters = new LinkedList<Parameter>(parameters);
        clone.attributes = new HashMap<String, StubObject>(attributes);
        clone.prune = prune;
        clone.uncompleted = uncompleted;
        clone.underscored = underscored;
        return clone;
    }
}
