/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.lib.jmi.mapping;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Iterator;
import java.util.List;
import org.netbeans.api.mdr.JMIStreamFactory;

/**
 *
 * @author  Martin Matula
 */
public class FileStreamFactory extends JMIStreamFactory {
    private final File targetDir;
    private final long modelLastModified;

    /** Creates a new instance of FileStreamFactory. Using this constructor is
     *  equivalent to using <code>FileStreamFactory(targetDir, null)</code>.
     *  @see #FileStreamFactory(File, java.util.Date)
     *  @param targetDir The target directory
     *  @throws IllegalArgumentException If <code>targetDir == null</code>,
     *          if the target directory does not already exist, or if the
     *          target file is not a directory.
     */
    public FileStreamFactory(File targetDir) {
        this(targetDir, null);
    }
    
    /** Creates a new instance of FileStreamFactory. FileStreamFactory will
     *  compare the last modified date of any existing target files against
     *  the value of the <code>modelLastModified</code> paramter. If the target
     *  file's last modified date is later than <code>modelLastModified/code>
     *  then {@link #createStream(List, String, String)} will return
     *  <code>null</code> to prevent the {@link JMIMapper} from overwriting
     *  the file. If <code>modelLastModified == null</code>, then
     *  <code>createStream</code> will never return <code>null</code>.
     *
     *  @param targetDir The target directory
     *  @param modelLastModified The date that the MOF model was last changed,
     *                           or <code>null</code>.
     *  @throws IllegalArgumentException If <code>targetDir == null</code>,
     *          if the target directory does not already exist, or if the
     *          target file is not a directory.
     */
    public FileStreamFactory(File targetDir, java.util.Date modelLastModified) {
        if (targetDir == null)
            throw new IllegalArgumentException("ERROR: targetDir is null");
        else if (!targetDir.exists())
            throw new IllegalArgumentException("ERROR: targetDir does not exist");
        else if (!targetDir.isDirectory()) {
            throw new IllegalArgumentException("ERROR: targetDir has to be a directory");
        }
        this.targetDir = targetDir;
        this.modelLastModified = modelLastModified != null ? modelLastModified.getTime() : 0L;
    }
    
    public OutputStream createStream(List pkg, String className, String extension) throws IOException {
        File current = targetDir;
        for (Iterator it = pkg.iterator(); it.hasNext();) {
            current = new File(current, (String) it.next());
            if (!current.exists()) {
                current.mkdir();
            }
        }
        return createStream(new File(current, className + "." + extension));
    }

    /** This method is called by {@link #createStream(List,String,String)} to
     *  really create the stream. It return creates the FileOutputSream
     *  corresponding to the given file if upToDate is false; otherwise it
     *  returns null. This is a convenient place for subclasses to hook
     *  into the stream creation process for logging or checking different
     *  criteria besides just the file's last modified date.
     */
    protected OutputStream createStream(java.io.File file) throws IOException {
        return modelLastModified == 0 || modelLastModified > file.lastModified()
            ? new FileOutputStream(file) : null;
    }
    
}
