/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.mdr.persistence.memoryimpl;

import java.util.*;
import org.netbeans.mdr.persistence.*;

public class TransactionLog {

    // operations codes
    private static final byte OP_NONE = 0;
    private static final byte OP_ADD = 1;
    private static final byte OP_REMOVE = 2;
    private static final byte OP_REPLACE = 3;

    private HashMap log = new HashMap ();
    private SinglevaluedIndex index;
    private boolean rollingBack = false;

    // ..........................................................................
    
    public TransactionLog (SinglevaluedIndex index) {
        this.index = index;
    }
    
    // ..........................................................................
    
    /**
     * Clears the transaction log.
     */
    public void clear () {
        log.clear ();
    }

    /**
     * Logs add operation.
     */
    public void logAdd (Object key) {
        if (rollingBack)
            return;
        Record rec = (Record) log.get(key);
        if (rec == null) {
            rec = new Record (OP_REMOVE, null);
            log.put (key, rec);
        } else {
            rec.opCode = OP_REPLACE;
        }
    }

    /**
     * Logs remove operation.
     */
    public void logRemove (Object key, Object value) {
        if (rollingBack)
            return;
        Record rec = (Record) log.get(key);
        if (rec == null) {
            rec = new Record (OP_ADD, value);
            log.put (key, rec);
        } else {
            log.remove (key);
        }
    }

    /**
     * Logs replace operation.
     */
    public void logReplace (Object key, Object oldValue) {
        if (rollingBack)
            return;
        Record rec = (Record) log.get(key);
        if (rec == null) {
            rec = new Record (OP_REPLACE, oldValue);
            log.put (key, rec);
        }
    }        

    public void logValue (Object key, Object value) {
        if (rollingBack)
            return;
        log.put (key, new Record (OP_NONE, value));
    }
    
    public boolean isLogged (Object key) {
        return log.get (key) != null;
    }
    
    public void setDirty (Object key) {
        if (rollingBack)
            return;
        Record rec = (Record) log.get(key);
        if (rec.opCode == OP_NONE)
            rec.opCode = OP_REPLACE;
    }        
    
    /**
     * Rolls back all changes performed over the index during the recorded transaction.
     */
    public void rollBack () throws StorageException {
        rollingBack = true;
        Iterator iter = log.keySet().iterator();
        while(iter.hasNext()) {
            Object key = iter.next();
            Record rec = (Record) log.get(key);
            if (rec.opCode == OP_NONE)
                continue;
            Object value;
            if (rec.value instanceof ValueLog) {
                value = ((ValueLog) rec.value).resolveOriginalValue ();
            } else {
                value = rec.value;
            }
            switch (rec.opCode) {
                case OP_ADD: 
                    index.add (key, value);
                break;
                case OP_REMOVE:
                    index.remove (key);
                break;
                case OP_REPLACE:
                    index.replace (key, value);
                break;
            } // switch                                    
        }
        rollingBack = false;
    }

    // ValueLog .................................................................
    
    public interface ValueLog {
        
        public Object resolveOriginalValue () throws StorageException;
        
    } // ValueLog
    
    // Record ...................................................................

    private static class Record {
        
        byte opCode; // operation code
        Object value;

        public Record (byte opCode, Object value) {
            this.opCode = opCode;
            this.value = value;
        }
        
    } // Record

}
