/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 */
package org.foo.wavutils;


import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.TargetDataLine;

import org.openide.ErrorManager;
import org.openide.util.RequestProcessor;

/**
 * Reads the audio data from a JavaSound input line to the PromptEditor.
 *
 * @author Rich Unger
 */
public class Recorder implements Runnable {

    public Recorder( AudioUI editor )
    {
        audioUI = editor;
    }

    public void start() {
        task = RequestProcessor.getDefault().post(this);
        task.addTaskListener(audioUI);
    }

    public void stop() {
        task = null;
    }

    private void shutDown()
    {
        SamplingGraph graph = audioUI.getSamplingGraph();
        graph.setCaptureLine(null);

        if (task != null) {
            task = null;
            graph.stop();
        }
    }

    public void run() {
        // define the required attributes for our line,
        // and make sure a compatible line is supported.

        DataLine.Info info = new DataLine.Info(TargetDataLine.class,
            Formats.INTERNAL_FORMAT);

        if (!AudioSystem.isLineSupported(info)) {
            ErrorManager.getDefault().log(ErrorManager.ERROR,
                    "Line matching " + info + " not supported.");
            shutDown();
            return;
        }

        // get and open the target data line for capture.

        try {
            dataLine = (TargetDataLine) AudioSystem.getLine(info);
            dataLine.open(Formats.INTERNAL_FORMAT, dataLine.getBufferSize());
        } catch (LineUnavailableException ex) {
            ErrorManager em = ErrorManager.getDefault();
            em.notify(em.annotate(ex, "Unable to open the line: " + ex.getMessage()));
            shutDown();
            return;
        } catch (SecurityException ex) {
            ErrorManager.getDefault().notify(ex);
            shutDown();
            return;
        } catch (Exception ex) {
            ErrorManager.getDefault().notify(ex);
            shutDown();
            return;
        }

        // record the audio data
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        int frameSizeInBytes = Formats.INTERNAL_FORMAT.getFrameSize();
        int bufferLengthInFrames = dataLine.getBufferSize() / 8;
        int bufferLengthInBytes = bufferLengthInFrames * frameSizeInBytes;
        byte[] data = new byte[bufferLengthInBytes];
        int numBytesRead;

        // link the line to the sampling graph
        audioUI.getSamplingGraph().setCaptureLine(dataLine);

        // start recording
        dataLine.start();

        while (task != null) {
            if((numBytesRead = dataLine.read(data, 0, bufferLengthInBytes)) == -1) {
                break;
            }
            out.write(data, 0, numBytesRead);
        }

        // we reached the end of the stream.  stop and close the line.
        dataLine.stop();
        dataLine.close();
        dataLine = null;

        // stop and close the output stream
        try {
            out.flush();
            out.close();
        } catch (IOException ex) {
            ex.printStackTrace();
        }

        // load bytes into the audio input stream for playback

        byte audioBytes[] = out.toByteArray();
        ByteArrayInputStream bais = new ByteArrayInputStream(audioBytes);
        AudioInputStream pcmStream = new AudioInputStream(bais,
                Formats.INTERNAL_FORMAT,
                audioBytes.length / frameSizeInBytes);

        WrappedAudioInputStream wrappedStream =
                new WrappedAudioInputStream( pcmStream );

        audioUI.setAudioInputStream( wrappedStream );
    }

    private TargetDataLine dataLine;
    private RequestProcessor.Task task;
    private final AudioUI audioUI;
}
