/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 */
package org.foo.wavutils;


import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.MouseEvent;
import java.awt.font.FontRenderContext;
import java.awt.font.LineBreakMeasurer;
import java.awt.font.TextAttribute;
import java.awt.font.TextLayout;
import java.awt.geom.Line2D;
import java.io.IOException;
import java.text.AttributedCharacterIterator;
import java.text.AttributedString;
import java.util.ArrayList;
import java.util.List;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.DataLine;
import javax.swing.JPanel;

import org.openide.ErrorManager;

/**
 * Render a WaveForm.
 *
 * This is a modified version of the sample code shipped with JavaSound,
 * written by Brian Lichtenwalter, and licensed royalty-free by Sun Microsystems.
 *
 * @author Rich Unger
 */

public class SamplingGraph extends JPanel implements Runnable
{
    //////////////////// CONSTRUCTORS /////////////////////////

    public SamplingGraph() {
        setBackground(Color.black);

        // mouse listeners for handling user selection
        SamplingGraph.MouseHandler mouseEventHandler =
                new SamplingGraph.MouseHandler();
        addMouseMotionListener(mouseEventHandler);
        addMouseListener(mouseEventHandler);

        // set the current selection to nothing
        m_iSelectionStart = m_iSelectionEnd = -1;
    }

    //////////////////// METHODS //////////////////////////////

    public void clearWaveForm()
    {
        audioData = null;
        m_sError = null;
        clearSelection();
        repaint();
    }

    public void createWaveForm( WrappedAudioInputStream audioInputStream )
    {
        if( audioInputStream != null && audioInputStream.getLength() > 0)
        {
            format = audioInputStream.getFormat();
            byte[] audioBytes = null;
            try
            {
                audioBytes = new byte[audioInputStream.getLength()];
                audioInputStream.read(audioBytes);
            }
            catch( Exception ex )
            {
                ErrorManager.getDefault().notify(ex);
                return;
            }

            createWaveForm(audioBytes);

            try
            {
                audioInputStream.reset();
            }
            catch( IOException ex )
            {
                ErrorManager.getDefault().notify(ex);
            }
        }
        repaint();
    }

    private void createWaveForm(byte[] audioBytes)
    {
        m_dDuration = audioBytes.length /
                      (double)format.getFrameSize() /
                      format.getSampleRate();
        // round it off to the thousandths
        m_dDuration = (double)Math.round(1000f * m_dDuration) / 1000f;

        audioData = null;  // clear the old audio data
        clearSelection();

        if (format.getSampleSizeInBits() == 16)
        {
             int iLengthInSamples = audioBytes.length / 2;
             audioData = new int[iLengthInSamples];

             if (format.isBigEndian())
             {
                 for (int i = 0; i < iLengthInSamples; i++)
                 {
                     /* First byte is MSB (high order) */
                     int MSB = audioBytes[2*i];
                     /* Second byte is LSB (low order) */
                     int LSB = audioBytes[2*i+1];
                     audioData[i] = MSB << 8 | (255 & LSB);
                 }
             }
             else
             {
                 for (int i = 0; i < iLengthInSamples; i++)
                 {
                     /* First byte is LSB (low order) */
                     int LSB = audioBytes[2*i];
                     /* Second byte is MSB (high order) */
                     int MSB = audioBytes[2*i+1];
                     audioData[i] = MSB << 8 | (255 & LSB);
                 }
             }
        }
        else if (format.getSampleSizeInBits() == 8)
        {
             int iLengthInSamples = audioBytes.length;
             audioData = new int[iLengthInSamples];
             if (format.getEncoding().toString().startsWith("PCM_SIGN"))
             {
                 for (int i = 0; i < audioBytes.length; i++)
                 {
                     audioData[i] = audioBytes[i];
                 }
             }
             else
             {
                 for (int i = 0; i < audioBytes.length; i++)
                 {
                     audioData[i] = audioBytes[i] - 128;
                 }
             }
        }
    }

    public boolean isRecording()
    {
        return m_bRecording;
    }

    public void setRecording(boolean b)
    {
        m_bRecording = b;
    }

    public void setCaptureLine(DataLine l)
    {
        m_captureLine = l;
    }

    public void setPlaybackLine(DataLine l)
    {
        m_playbackLine = l;
    }

    public String getErrorString()
    {
        return m_sError;
    }

    public int getSelectionStart()
    {
        return m_iSelectionStart;
    }

    public int getSelectionEnd()
    {
        return m_iSelectionEnd;
    }

    public void clearSelection()
    {
        m_iSelectionStart = m_iSelectionEnd = -1;
    }


    public void paint(Graphics g)
    {
        Dimension d = getSize();
        int w = d.width;
        int h = d.height;
        int INFOPAD = 15;
        
        List<Line2D> lines = new ArrayList<Line2D>();
        if (audioData != null && format != null)
        {

            int frames_per_pixel = (int)(m_dDuration * format.getSampleRate()) / w;
            byte my_byte = 0;
            double y_last = 0;
            for( double x=0; x < w; x++ )
            {
                int idx = (int) (frames_per_pixel * format.getChannels() * x);
                if( format.getSampleSizeInBits() == 8)
                {
                    my_byte = (byte) audioData[idx];
                }
                else
                {
                    my_byte = (byte) (128 * audioData[idx] / 32768);
                }
                double y_new = (h * (128 - my_byte) / 256);
                lines.add(new Line2D.Double(x, y_last, x, y_new));
                y_last = y_new;
            }
        }

        Graphics2D g2 = (Graphics2D) g;
        g2.setBackground(getBackground());
        g2.clearRect(0, 0, w, h);
        g2.setColor(Color.white);
        g2.fillRect(0, h-INFOPAD, w, INFOPAD);

        if (m_sError != null)
        {
            // Write out an error string instead of the wave form

            g2.setColor(s_jfcBlue);
            g2.setFont(new Font("serif", Font.BOLD, 18));
            g2.drawString("ERROR", 5, 20);
            AttributedString as = new AttributedString(m_sError);
            as.addAttribute(TextAttribute.FONT, s_font12, 0, m_sError.length());
            AttributedCharacterIterator aci = as.getIterator();
            FontRenderContext frc = g2.getFontRenderContext();
            LineBreakMeasurer lbm = new LineBreakMeasurer(aci, frc);
            float x = 5, y = 25;
            lbm.setPosition(0);
            while (lbm.getPosition() < m_sError.length()) {
                TextLayout tl = lbm.nextLayout(w-x-5);
                if (!tl.isLeftToRight()) {
                    x = w - tl.getAdvance();
                }
                tl.draw(g2, x, y += tl.getAscent());
                y += tl.getDescent() + tl.getLeading();
            }
        }
        else if (isRecording())
        {
            // while recording, just write out the current length in big
            // characters in the middle of the graph
            g2.setColor(s_jfcBlue);
            g2.setFont(new Font("serif", Font.BOLD, 36));
            g2.drawString("Length: " + String.valueOf(m_dSeconds), 25, 50);
        }
        else
        {
            // write the length and position at the bottom

            g2.setColor(Color.black);
            g2.setFont(s_font12);
            g2.drawString("Length: " + String.valueOf(m_dDuration) + "  Position: " + String.valueOf(m_dSeconds), 3, h-4);

            if (!lines.isEmpty()) {
                // .. render sampling graph ..
                g2.setColor(s_jfcBlue);
                for (int i = 1; i < lines.size(); i++) {
                    g2.draw((Line2D) lines.get(i));
                }

                // .. draw current position ..
                if (m_dSeconds != 0) {
                    double loc = m_dSeconds/m_dDuration*w;
                    // if we're just playing a selection, we need to move to
                    // the offset
                    if( m_iSelectionStart != -1 )
                    {
                        loc += Math.min( m_iSelectionStart, m_iSelectionEnd );
                    }
                    g2.setColor(s_pink);
                    g2.setStroke(new BasicStroke(3));
                    g2.draw(new Line2D.Double(loc, 0, loc, h-INFOPAD-2));
                }
            }

            // draw the current selection
            if( m_iSelectionStart != -1 )
            {
                g2.setXORMode(Color.yellow);
                g2.fillRect( Math.min(m_iSelectionStart, m_iSelectionEnd),
                             0,
                             Math.abs(m_iSelectionStart - m_iSelectionEnd),
                             getHeight());
            }
        }
    }

    public void start()
    {
        m_thread = new Thread(this);
        m_thread.setName("SamplingGraph");
        m_thread.start();
    }

    public void stop()
    {
        if (m_thread != null)
        {
            m_thread.interrupt();
        }
        m_thread = null;
    }

    /**
     * Runs a line across the sample waveform during playback.
     */
    public void run()
    {
        m_dSeconds = 0;
        while (m_thread != null)
        {
            if ((m_playbackLine != null) && (m_playbackLine.isOpen()) )
            {
                long milliseconds = (m_playbackLine.getMicrosecondPosition() / 1000);
                m_dSeconds =  milliseconds / 1000.0;
            }
            else if ( (m_captureLine != null) && (m_captureLine.isActive()) )
            {
                long milliseconds = (m_captureLine.getMicrosecondPosition() / 1000);
                m_dSeconds =  milliseconds / 1000.0;
            }

            try { Thread.sleep(100); } catch (Exception e) { break; }

            repaint();

            while ((m_captureLine != null && !m_captureLine.isActive()) ||
                   (m_playbackLine != null && !m_playbackLine.isOpen()))
            {
                try { Thread.sleep(10); } catch (Exception e) { break; }
            }
        }
        m_dSeconds = 0;
        repaint();
    }

    ////////////////// INNER CLASSES ///////////////////////////
    class MouseHandler extends javax.swing.event.MouseInputAdapter
    {
        public void mouseDragged( MouseEvent evt )
        {
            // Set the current selection
            if( m_iSelectionStart == -1 )
            {
                m_iSelectionStart = m_iSelectionEnd = evt.getX();
            }
            else
            {
                int x = evt.getX();
                if( x >= 0 && x <= SamplingGraph.this.getWidth() )
                {
                    m_iSelectionEnd = x;
                }
            }

            SamplingGraph.this.repaint();
        }

        public void mousePressed( MouseEvent evt )
        {
            // clear the current selection
            SamplingGraph.this.clearSelection();

            SamplingGraph.this.repaint();
        }
    }

    //////////////////// INSTANCE VARIABLES ////////////////////
    private Thread m_thread;

    private double m_dDuration, m_dSeconds;
    private boolean m_bRecording = false;
    private DataLine m_captureLine, m_playbackLine;
    private String m_sError;
    private int m_iSelectionStart;
    private int m_iSelectionEnd;
    private int[] audioData;
    private AudioFormat format;


    //////////////////// CONSTANTS //////////////////////////////
    private static final Font s_font12 = new Font("serif", Font.PLAIN, 12);
    private static final Color s_jfcBlue = new Color(204, 204, 255);
    private static final Color s_pink = new Color(255, 175, 175);

}
