/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2008 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2008 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.examples.cards.api;

import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;
import org.netbeans.examples.cards.api.Pile.VisibilityPolicy;
import org.openide.util.ImageUtilities;

/**
 *
 * @author Tim Boudreau
 */
public final class Deck {

    private final List<Card> cards;
    private final Game game;

    public Deck(int jokerCount) {
        this(new Game.DefaultGame(jokerCount));
    }

    public Deck(Game game) {
        this.game = game;
        cards = initCards(game.getJokerCount());
    }

    public List<Card> getCards() {
        return Collections.<Card>unmodifiableList(cards);
    }
    Suit spades, clubs, diamonds, hearts;

    private List<Card> initCards(int jokerCount) {
        List<Card> theCards = new ArrayList<Card>();
        spades = new Suit(Suits.SPADES, this);
        clubs = new Suit(Suits.CLUBS, this);
        diamonds = new Suit(Suits.DIAMONDS, this);
        hearts = new Suit(Suits.HEARTS, this);
        theCards.addAll(spades.getCards());
        theCards.addAll(clubs.getCards());
        theCards.addAll(diamonds.getCards());
        theCards.addAll(hearts.getCards());
        for (int i = 0; i < jokerCount; i++) {
            theCards.add(new Card(Rank.JOKER, null, this));
        }
        return theCards;
    }

    public void shuffle() {
        Collections.shuffle(cards);
    }

    public Pile createPile(Pile.VisibilityPolicy policy, Collection<Card> cards) {
        return new Pile(policy, new ArrayList<Card>(cards));
    }

    /**
     * Deal hands of the specified size into the passed list and return
     * those cards that were not dealt out.  Note that this will remove
     * all cards from any existing Pile objects created by this Deck.
     */
    public Pile deal(List<Pile> hands, int count, int handsize, VisibilityPolicy policyForRemainder) {
        shuffle();
        int ix = 0;
        for (int i = 0; i < count; i++) {
            List<Card> subl = cards.subList(ix, ix + handsize);
            Pile pile = new Pile(VisibilityPolicy.ALL_VISIBLE, subl);
            hands.add(pile);
            ix += handsize;
        }
        List<Card> remainingCards = cards.subList(handsize * count, cards.size() - 1);
        Pile remaining = new Pile(policyForRemainder, remainingCards);
        return remaining;
    }

    @Override
    public String toString() {
        return super.toString();
    }

    public int size() {
        return cards.size();
    }
    private BufferedImage back = null;

    public BufferedImage getFaceDownCardImage() {
        if (back == null) {
            back = loadRandomBack();
        }
        return back;
    }

    public boolean contains(Card card) {
        return cards.contains(card);
    }

    private static BufferedImage loadRandomBack() {
        Random r = new Random(System.currentTimeMillis());
        StringBuilder sb = new StringBuilder();
        sb.append("org/netbeans/examples/cards/resources/back-");
        if (r.nextBoolean()) {
            sb.append("red");
        } else {
            sb.append("blue");
        }
        sb.append("-150-");
        sb.append(r.nextInt(3) + 1);
        sb.append(".png");
        return (BufferedImage) ImageUtilities.loadImage(sb.toString());
    }

    Card getCard(Rank id, Suits suit) {
        switch (suit) {
            case HEARTS:
                return hearts.getCard(id);
            case SPADES:
                return spades.getCard(id);
            case CLUBS:
                return clubs.getCard(id);
            case DIAMONDS:
                return diamonds.getCard(id);
            default:
                throw new AssertionError();
        }
    }

    public List<Pile> allPiles() {
        Set<Pile> piles = new HashSet<Pile>();
        for (Card card : cards) {
            Pile pile = card.getPile();
            if (pile != null) {
                piles.add(card.getPile());
            }
        }
        List<Pile> result = new ArrayList<Pile>(piles);
        Collections.sort(result, new PileNameComparator());
        return Collections.<Pile>unmodifiableList(result);
    }

    public Game getGame() {
        return game;
    }

    public List<Card> getUndealt() {
        List<Card> result = new ArrayList<Card>(cards.size());
        for (Card card : cards) {
            if (card.getPile() == null) {
                result.add(card);
            }
        }
        return result;
    }

    private class PileNameComparator implements Comparator<Pile> {

        public int compare(Pile o1, Pile o2) {
            return game.nameOf(o1).compareToIgnoreCase(game.nameOf(o2));
        }
    }
}
