package org.netbeans.modules.manifesteditor.editor;

import java.lang.reflect.InvocationTargetException;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.jar.Attributes;
import java.util.jar.Manifest;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import org.openide.nodes.Node.Property;
import org.openide.nodes.PropertySupport;
import org.openide.nodes.Sheet;
import org.openide.util.NbBundle;

final class ManifestDataNode extends AbstractNode {
    
    private Attributes attrs;
    private boolean sheet;
    private final ChangeCallback callback;
    
    private ManifestDataNode(String name, Attributes attrs, ChangeCallback callback) {
        super(Children.LEAF);
        setName(name);
        
        if ("Main".equals(name)) {
            setDisplayName(NbBundle.getMessage(ManifestDataNode.class, "CTL_MainAttributes"));
            setIconBaseWithExtension("org/netbeans/modules/manifesteditor/manifest.png");
        } else {
            setDisplayName(NbBundle.getMessage(ManifestDataNode.class, "CTL_Attributes", name));
            setIconBaseWithExtension("org/netbeans/modules/manifesteditor/section.png");
        }
        
        this.attrs = attrs;
        this.callback = callback;
    }
    
    public static interface ChangeCallback {
        public void change(String section, String name, String oldValue, String newValue)
        throws IllegalArgumentException;
    }
    
    public static Node createManifestModel(Manifest mf, ChangeCallback callback) {
        return new AbstractNode(new Entries(attributes(mf), callback));
    }
    
    private static Map<String, Attributes> attributes(final Manifest mf) {
        Map<String, Attributes> en;
        en = new LinkedHashMap<String, Attributes>();
        en.put("Main", mf.getMainAttributes());
        en.putAll(mf.getEntries());
        return en;
    }
    
    public static void refresh(Node n, Manifest mf) {
        assert n.getChildren() instanceof Entries;
        
        Entries e = (Entries)n.getChildren();
        e.refresh(attributes(mf));
    }
    
    protected synchronized Sheet createSheet() {
        this.sheet = true;
        
        Sheet sheet = new Sheet();
        Sheet.Set ps = new Sheet.Set();
        ps.setName("Attributes"); // NOI18N
        ps.setDisplayName(NbBundle.getMessage(ManifestDataNode.class, "PROP_Attributes"));
        ps.setShortDescription(NbBundle.getMessage(ManifestDataNode.class, "HINT_Attributes"));
        
        
        refreshSheet(ps);
        
        sheet.put(ps);
        
        return sheet;
    }
    
    public synchronized void refresh(Attributes attrs) {
        this.attrs = attrs;
        refreshSheet(getSheet().get("Attributes"));
    }
    
    private void refreshSheet(Sheet.Set set) {
        assert Thread.holdsLock(this);
        
        Property[] clone = (Property[])set.getProperties().clone();
        for (Property p : set.getProperties()) {
            if (!attrs.containsKey(p.getName())) {
                set.remove(p.getName());
            }
        }
        
        for (Object name : attrs.keySet()) {
            Prop p = new Prop(((Attributes.Name)name).toString());
            set.put(p);
        }
    }
    
    
    private static class Entries extends Children.Keys {
        private java.util.Map<String,Attributes> entries;
        private ChangeCallback callback;
        
        public Entries(java.util.Map<String,Attributes> entries, ChangeCallback callback) {
            this.entries = entries;
            this.callback = callback;
            setKeys(entries.keySet());
        }
        
        protected Node[] createNodes(Object okey) {
            String key = (String)okey;
            return new Node[] { new ManifestDataNode(key, entries.get(key), callback) };
        }
        
        public void refresh(java.util.Map<String,Attributes> entries) {
            this.entries = entries;
            setKeys(entries.keySet());
            
            if (isInitialized()) {
                for (Node n : getNodes()) {
                    ManifestDataNode m = (ManifestDataNode)n;
                    m.refresh(entries.get(m.getName()));
                }
            }
        }
    }
    
    private class Prop extends PropertySupport {
        public Prop(String name) {
            super(name, String.class, name, name, true, true);
        }
        
        
        public Object getValue() throws IllegalAccessException, InvocationTargetException {
            return attrs.getValue(getName());
        }
        
        public void setValue(Object val) throws IllegalAccessException, IllegalArgumentException, InvocationTargetException {
            callback.change(
                    ManifestDataNode.this.getName(),
                    getName(),
                    (String)getValue(),
                    (String)val
                    );
        }
    }
    
}
