/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2008 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2008 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.examples.manifestsupport;

import java.awt.EventQueue;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.nio.charset.Charset;
import junit.framework.*;
import org.netbeans.examples.manifestsupport.api.ManifestProvider;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.filesystems.LocalFileSystem;
import org.openide.loaders.DataObject;

/**
 *
 * @author Tim Boudreau
 */
public class ManifestDataObjectTest extends TestCase {

    public ManifestDataObjectTest(String testName) {
        super(testName);
    }
    private static final String BAD_MANIFEST_CONTENT =
            "Manifest-Version: 1.0\n" +
            "junk junk junk\n" +
            "some more junk\n";
    private static final String GOOD_MANIFEST_CONTENT =
            "Manifest-Version: 1.0\n" +
            "Java-Bean: true\n" +
            "OpenIDE-Module-Name: com.foo.bar";
    private FileObject bad;
    private FileObject good;

    protected void setUp() throws Exception {
        //Get the system temp dir
        File dir = new File(System.getProperty("java.io.tmpdir"));
        File goodManifest = new File(dir, System.currentTimeMillis() +
                "good.mf");

        goodManifest.createNewFile();
        writeFile(GOOD_MANIFEST_CONTENT, goodManifest);

        File badManifest = new File(dir, System.currentTimeMillis() +
                "bad.mf");
        badManifest.createNewFile();
        writeFile(BAD_MANIFEST_CONTENT, badManifest);

        //Deprecated but only way to do it for tests w/o loading
        //all of core to get declarative mime resolvers loaded.  This
        //call just manually sets the mime type to be associated with
        //the .mf extension
        FileUtil.setMIMEType("mf", "text/x-manifest"); // NOI18N

        //Create a local filesystem so we can make FileObjects from Files
        LocalFileSystem fs = new LocalFileSystem();
        fs.setRootDirectory(dir);

        good = fs.getRoot().getFileObject(goodManifest.getName());
        bad = fs.getRoot().getFileObject(badManifest.getName());

        //Sanity check
        assertNotNull(good);
        assertNotNull(bad);
    }

    private void writeFile(String content, File file) throws Exception {
        Charset charset = Charset.forName("UTF-8");
        ByteBuffer buf = charset.encode(content);
        FileOutputStream str = new FileOutputStream(file);
        FileChannel channel = str.getChannel();
        FileLock lock = channel.lock();
        try {
            channel.write(buf);
            channel.force(true);
        } finally {
            lock.release();
            channel.close();
        }
    }

    protected void tearDown() throws Exception {
        good.delete();
        bad.delete();
    }

    public void testManifestProviderPresent() throws Exception {
        System.out.println("testManifestProviderPresent");
        DataObject goodDob = DataObject.find(good);
        assertTrue(goodDob instanceof ManifestDataObject);
        assertNotNull(goodDob.getCookie(ManifestProvider.class));
    }

    public void testManifestProviderThrowsExceptionWhenCalledInEventThread()
            throws Exception {
        System.out.println("testManifestProviderPresent");
        DataObject goodDob = DataObject.find(good);
        final ManifestProvider provider = (ManifestProvider) goodDob.getCookie(ManifestProvider.class);
        class R implements Runnable {

            private Exception exception;

            public void run() {
                try {
                    provider.getManifest();
                } catch (Exception e) {
                    exception = e;
                }
            }
        }
        ;
        R r = new R();
        EventQueue.invokeAndWait(r);
        assertNotNull(r.exception);
    }

    public void testDataObject() throws Exception {
        System.out.println("testDataObject");
        DataObject goodDob = DataObject.find(good);
        DataObject badDob = DataObject.find(bad);
        assertTrue(goodDob instanceof ManifestDataObject);
        assertTrue(badDob instanceof ManifestDataObject);
        ManifestProvider mp = (ManifestProvider) goodDob.getCookie(ManifestProvider.class);
        assertNotNull(mp);
        assertNotNull(mp.getManifest());
        //Prove that you can get the ManifestProvider all the ways you should
        //be able to
        assertSame(mp, goodDob.getNodeDelegate().getCookie(
                ManifestProvider.class));

        assertSame(mp, goodDob.getNodeDelegate().getLookup().lookup(
                ManifestProvider.class));
    }

    public void testBadManifest() throws Exception {
        System.out.println("testBadManifest");
        DataObject badDob = DataObject.find(bad);
        ManifestProvider mp = (ManifestProvider) badDob.getCookie(ManifestProvider.class);
        assertNotNull(mp);
        assertNull(mp.getManifest());
    }

    public void testFileWritesChangeResult() throws Exception {
        System.out.println("testFileWritesChangeResult");
        DataObject goodDob = DataObject.find(good);
        ManifestProvider mp = (ManifestProvider) goodDob.getCookie(ManifestProvider.class);
        assertNotNull(mp.getManifest());
        writeFileObject(BAD_MANIFEST_CONTENT, good);
        assertNull(mp.getManifest());
        writeFileObject(GOOD_MANIFEST_CONTENT, good);
        assertNotNull(mp.getManifest());
    }
    /*
    public void testIconChanges() throws Exception {
    System.out.println("testIconChanges");

    DataObject goodDob = DataObject.find (good);
    Node n = goodDob.getNodeDelegate();
    assertTrue (n instanceof ManifestDataNode);
    PCL pcl = new PCL();
    n.addPropertyChangeListener(pcl);

    Image icon = n.getIcon (BeanInfo.ICON_COLOR_16x16);
    writeFileObject (BAD_MANIFEST_CONTENT, good);
    pcl.waitForEvent();
    pcl.assertFired();

    Image icon2 = n.getIcon (BeanInfo.ICON_COLOR_16x16);
    assertNotSame (icon, icon2);
    BufferedImage bufimg = toBufferedImage (icon2);

    Color px = new Color (bufimg.getRGB (14, 14));
    assertEquals (Color.RED, px);

    writeFileObject (GOOD_MANIFEST_CONTENT, good);
    pcl.waitForEvent();
    pcl.assertFired();

    Image icon3 = n.getIcon (BeanInfo.ICON_COLOR_16x16);
    assertSame (icon, icon3);
    }

    private BufferedImage toBufferedImage (Image img) {
    if (img instanceof BufferedImage) {
    return ((BufferedImage) img);
    } else {
    BufferedImage result = new BufferedImage (img.getWidth(null),
    img.getHeight(null), BufferedImage.TYPE_INT_ARGB);
    Graphics2D g = result.createGraphics();
    g.drawImage(img, 0, 0, null);
    return result;
    }
    }

    private static class PCL implements PropertyChangeListener {

    public void waitForEvent () {
    try {
    synchronized (this) {
    wait (10000);
    }
    } catch (Exception e) {

    }
    }

    public void assertNotFired() {
    assertNull (evt);
    }

    private PropertyChangeEvent evt;
    public PropertyChangeEvent assertFired() {
    PropertyChangeEvent result = evt;
    evt = null;
    assertNotNull (result);
    return result;
    }

    public void propertyChange (PropertyChangeEvent pce) {
    System.err.println("PROPERTY CHANGE " + pce);
    if (Node.PROP_ICON.equals(pce.getPropertyName())) {
    evt = pce;
    synchronized (this) {
    notifyAll();
    }
    }
    }
    }
     */

    private void writeFileObject(String content, FileObject file) throws Exception {
        org.openide.filesystems.FileLock lock = file.lock();
        OutputStream out = file.getOutputStream(lock);
        byte[] bytes = content.getBytes("UTF-8");
        ByteArrayInputStream in = new ByteArrayInputStream(bytes);
        try {
            FileUtil.copy(in, out);
        } finally {
            out.close();
            lock.releaseLock();
        }
    }
}
