/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2008 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2008 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.examples.xmlpseudonavigator;

import java.awt.EventQueue;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.swing.ListModel;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;
import org.openide.filesystems.FileChangeAdapter;
import org.openide.filesystems.FileChangeListener;
import org.openide.filesystems.FileEvent;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.util.RequestProcessor;

/**
 * A ListModel to represent the tags in an XML or HTML file
 *
 * @author Tim Boudreau
 */
public class XmlFileListModel extends FileChangeAdapter implements ListModel,
        Runnable, FileChangeListener {
    //The fileobject whose content we are modelling

    final FileObject file;
    //Our listener list
    private final List<ListDataListener> listeners =
            Collections.synchronizedList(new LinkedList<ListDataListener>());
    //Flag indicating there is a pending job to do a parse
    private volatile boolean enqueued = false;
    //Regexp we'll use to match things - it will skip very short tags and 
    //only id the first tag on a line
    private static Pattern pattern = Pattern.compile("<(\\w\\w\\w.*?)>.*?\n"); //NOI18N
    //Our model content list, which will only be accessed on the event thread
    private final List<XmlListModelItem> contents =
            new ArrayList<XmlListModelItem>(20);
    //A list we'll use to populate content, doing our I/O outside the event 
    //thread
    private List<XmlListModelItem> pendingContent = null;

    public XmlFileListModel(FileObject file) {
        this.file = file;
        file.addFileChangeListener(FileUtil.weakFileChangeListener(this, file));
        enqueue();
    }

    FileObject getFile() {
        return file;
    }

    public int getSize() {
        return contents.size();
    }

    public Object getElementAt(int index) {
        return contents.get(index);
    }

    public void addListDataListener(ListDataListener l) {
        listeners.add(l);
    }

    public void removeListDataListener(ListDataListener l) {
        listeners.remove(l);
    }

    private void enqueue() {
        if (!enqueued) {
            RequestProcessor.getDefault().post(this);
        }
    }

    //From FileChangeListener - will be called if the file is saved
    public void fileChanged(FileEvent evt) {
        enqueue();
    }

    //From FileChangeListener - will be called if the file is deleted
    public void fileDeleted(FileEvent evt) {
        enqueue();
    }

    public void run() {
        //We run first off the event thread, and then then 
        //invokeLater ourselves to call this method again from the event
        //thread once we are done
        if (!EventQueue.isDispatchThread()) {
            try {
                //Make the list of our content on the background thread
                pendingContent = parseFileAndCreateItems();
            } finally {
                EventQueue.invokeLater(this);
            }
        } else {
            //Second invocation, we are on the event queue now - copy
            //the list over
            try {
                assert pendingContent != null;
                int sz = Math.max(pendingContent.size(), contents.size());

                //Empty the backing storage.  We're running on the event
                //thread now, so nobody should be calling ListModel methods 
                //while we do this
                contents.clear();
                //Replace the former contents with the new list of items
                contents.addAll(pendingContent);

                ListDataListener[] l =
                        listeners.toArray(new ListDataListener[0]);

                ListDataEvent e = new ListDataEvent(this,
                        ListDataEvent.CONTENTS_CHANGED, 0, sz);

                for (int i = 0; i < l.length; i++) {
                    l[i].contentsChanged(e);
                }
            } finally {
                pendingContent = null;
                enqueued = false;
            }
        }
    }

    private List<XmlListModelItem> parseFileAndCreateItems() {
        List<XmlListModelItem> result = new LinkedList<XmlListModelItem>();
        if (file.isValid() && file.getSize() < Integer.MAX_VALUE) {
            try {
                String content = readFile();
                Matcher matcher = pattern.matcher(content);
                while (matcher.find()) {
                    result.add(new XmlListModelItem(this,
                            matcher.group(1), matcher.start()));
                }
            } catch (IOException ex) {
                Logger.getLogger("global").log(
                        Level.INFO, ex.getMessage(), ex);
            }
        }
        return result;
    }

    private String readFile() throws IOException {
        InputStream in = file.getInputStream();
        ByteArrayOutputStream bytes = new ByteArrayOutputStream((int) file.getSize());
        try {
            FileUtil.copy(in, bytes);
            byte[] arr = bytes.toByteArray();
            String s = new String(arr, 0, arr.length,
                    Charset.defaultCharset());

            String lineSep = System.getProperty("line.separator");
            //We are computing offsets into a Swing document, where
            //the line separator will always be \n no matter what the
            //OS does
            if (lineSep.length() > 1) {
                s = s.replace(lineSep, "\n");
            }
            return s;
        } finally {
            in.close();
        }
    }
}
