/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.examples.modules.povproject;

import java.awt.EventQueue;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.Iterator;
import java.util.Properties;
import java.util.prefs.Preferences;
import javax.swing.JFileChooser;
import org.netbeans.api.project.ProjectInformation;
import org.netbeans.modules.examples.api.povray.MainFileProvider;
import org.netbeans.modules.examples.api.povray.ViewService;
import org.openide.DialogDisplayer;
import org.openide.ErrorManager;
import org.openide.NotifyDescriptor;
import org.openide.awt.StatusDisplayer;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.util.NbBundle;
import org.openide.util.RequestProcessor;
import org.openide.windows.IOProvider;
import org.openide.windows.InputOutput;
import org.openide.windows.OutputWriter;
import org.openide.windows.WindowManager;

/**
 * Encapsulates a POV-Ray executable.
 *
 * @author Tim Boudreau
 */
class Povray {
    private static File povray = null;
    private static File include = null;

    /** Preferences key for the povray executable */
    private static final String KEY_POVRAY_EXEC = "povray";
    /** Preferences key for the povray standard includes dir */
    private static final String KEY_POVRAY_INCLUDES = "include";

    private final RendererServiceImpl renderService;
    private final FileObject toRender;
    private final Properties settings;

    Povray (RendererServiceImpl renderService, FileObject toRender, Properties settings) {
        this.renderService = renderService;
        this.toRender = toRender;
        this.settings = settings == null ? renderService.getRendererSettings(
                renderService.getPreferredRendererSettingsName()) : settings;
    }

    private static File getPovray() {
        if (povray == null || !povray.exists()) {
            Preferences prefs = RendererServiceImpl.getPreferences();
            String loc = prefs.get(KEY_POVRAY_EXEC, null);
            if (loc != null) {
                povray = new File (loc);
            }

            if (povray == null || !povray.exists()) {
                File maybePov = locate("TTL_FindPovray");

                if (maybePov != null && maybePov.getPath().endsWith("pvengine.exe")) {
                    //Warn the user to get a command line build
                    NotifyDescriptor msg = new NotifyDescriptor.Confirmation (
                        NbBundle.getMessage(RendererServiceImpl.class,
                            "MSG_WindowsWarning"),
                        NotifyDescriptor.WARNING_MESSAGE);

                    Object result = DialogDisplayer.getDefault().notify(msg);
                    if (result == NotifyDescriptor.CANCEL_OPTION) {
                        return null;
                    }
                }
                povray = maybePov;
                if (povray != null) {
                    prefs.put (KEY_POVRAY_EXEC, povray.getPath());
                }
            }
        }
        return povray;
    }


    private static File getStandardIncludeDir (File povray) {
        if (include != null) {
            return include;
        }
        Preferences prefs = RendererServiceImpl.getPreferences();
        String loc = prefs.get(KEY_POVRAY_INCLUDES, null);
        if (loc != null) {
            include = new File (loc);
            if (!include.exists()) {
                include = null;
            }
        }
        if (include == null) {
            include = new File (povray.getParentFile().getParent() +
                    File.separator + "include");

            if (!include.exists()) {
                include = locate ("TTL_FindIncludes");
                if (include != null) {
                    prefs.put(KEY_POVRAY_INCLUDES, include.getPath());
                } else {
                    include = null;
                }
            }
        }
        return include;
    }

    private static File locate(String key) {
        JFileChooser jfc = new JFileChooser();
        jfc.setDialogTitle(NbBundle.getMessage(Povray.class, key));
        jfc.setFileSelectionMode (JFileChooser.FILES_ONLY);
        jfc.showOpenDialog(null);
        File result = jfc.getSelectedFile();
        return result;
    }

    private File getFileToRender() throws IOException {
        FileObject render = toRender;
        if (render == null) {
            PovrayProject proj = renderService.getProject();
            MainFileProvider provider = 
                proj.getLookup().lookup (MainFileProvider.class);
            if (provider == null) {
                throw new IllegalStateException ("Main file provider missing");
            }
            render = provider.getMainFile();
            if (render == null) {
                ProjectInformation info = 
                        proj.getLookup().lookup(ProjectInformation.class);

                //XXX let the user choose
                throw new IOException (NbBundle.getMessage(Povray.class,
                        "MSG_NoMainFile", info.getDisplayName()));
            }
        }
        assert render != null;
        File result = FileUtil.toFile (render);
        if (result == null) {
                throw new IOException (NbBundle.getMessage(Povray.class,
                        "MSG_VirtualFile", render.getName()));
        }
        assert result.exists();
        assert result.isFile();
        return result;
    }

    private String getCmdLineArgs(File includesDir) {
        StringBuffer cmdline = new StringBuffer();
        for (Iterator i=settings.keySet().iterator(); i.hasNext();) {
            String key = (String) i.next();
            String val = settings.getProperty(key);
            cmdline.append ('+');
            cmdline.append (key);
            cmdline.append (val);
            cmdline.append (' ');
        }
        cmdline.append ("+L");
        cmdline.append (includesDir.getPath());
        return cmdline.toString();
    }

    private File getImagesDir() {
        PovrayProject proj = renderService.getProject();
        FileObject fob = proj.getImagesFolder(true);
        File result = FileUtil.toFile(fob);
        assert result != null && result.exists();
        return result;
    }

    private String stripExtension(File f) {
        String sceneName = f.getName();
        int endIndex;
        if ((endIndex = sceneName.lastIndexOf('.')) != -1) {
            sceneName = sceneName.substring(0, endIndex);
        }
        return sceneName;
    }

    private void showMsg (String msg) {
        StatusDisplayer.getDefault().setStatusText(msg);
    }

    public FileObject render () throws IOException {
        if (EventQueue.isDispatchThread()) {
            throw new IllegalStateException ("Tried to run povray from the " +
                    "event thread");
        }

        //Find the scene file pass to POV-Ray as a java.io.File
        File scene;
        try {
            scene = getFileToRender();
        } catch (IOException ioe) {
            showMsg (ioe.getMessage());
            return null;
        }

        //Get the POV-Ray executable
        File povray = getPovray();
        if (povray == null) {
            //The user cancelled the file chooser w/o selecting
            showMsg(NbBundle.getMessage(Povray.class, "MSG_NoPovrayExe"));
            return null;
        }

        //Get the include dir, if it isn't under povray's home dir
        File includesDir = getStandardIncludeDir(povray);
        if (includesDir == null) {
            //The user cancelled the file chooser w/o selecting
            showMsg (NbBundle.getMessage(Povray.class, "MSG_NoPovrayInc"));
            return null;
        }

        //Find the image output directory for the project
        File imagesDir = getImagesDir();

        //Assemble and format the line switches for the POV-Ray process based
        //on the contents of the Properties object
        String args = getCmdLineArgs(includesDir);
        String outFileName = stripExtension (scene) + ".png";

        //Compute the name of the output image file
        File outFile = new File(imagesDir.getPath() + File.separator +
                outFileName);

        //Delete it if it exists, so that any current tab viewing the file is
        //closed and the file will definitely be re-read when it is re-opened
        if (outFile.exists() && !outFile.delete()) {
            showMsg (NbBundle.getMessage(Povray.class,
                    "LBL_CantDelete", outFile.getName()));
            return null;
        }

        //Append the input file and output file arguments to the command line
        String cmdline = povray.getPath() + ' ' + args + " +I" +
                scene.getPath() + " +O" + outFile.getPath();

        System.err.println(cmdline);

        showMsg (NbBundle.getMessage(Povray.class, "MSG_Rendering",
                scene.getName()));
        final Process process = Runtime.getRuntime().exec (cmdline);

        //Get the standard out
        InputStream out = new BufferedInputStream (process.getInputStream(), 8192);
        //Get the standard in
        InputStream err = new BufferedInputStream (process.getErrorStream(), 8192);

        //Create readers for each
        final Reader outReader = new BufferedReader (new InputStreamReader (out));
        final Reader errReader = new BufferedReader (new InputStreamReader (err));

        InputOutput io = IOProvider.getDefault().getIO(scene.getName(), false);

        //Force it to open the output window/activate our tab
        io.select();

        ////Print the command line we're calling for debug purposes
        io.getOut().println(cmdline);

        //Create runnables to poll each output stream
        OutHandler processSystemOut = new OutHandler (outReader, io.getOut());
        OutHandler processSystemErr = new OutHandler (errReader, io.getErr());

        //Get two different threads listening on the output & err
        RequestProcessor.getDefault().post(processSystemOut);
        RequestProcessor.getDefault().post(processSystemErr);

        try {
            //Hang this thread until the process exits
            process.waitFor();
        } catch (InterruptedException ex) {
            ErrorManager.getDefault().notify(ex);
        }

        //Close the output window's streams (title will become non-bold)
        processSystemOut.close();
        processSystemErr.close();

        if (outFile.exists() && process.exitValue() == 0) {
            //Try to find the new image file
            FileObject outFileObject = FileUtil.toFileObject(outFile);
            showMsg (NbBundle.getMessage(Povray.class, "MSG_Success",
                outFile.getPath()));
            return outFileObject;
        } else {
            showMsg (NbBundle.getMessage(Povray.class, "MSG_Failure",
                scene.getPath()));
            return null;
        }
    }

    static class OutHandler implements Runnable {
        private Reader out;
        private OutputWriter writer;
        public OutHandler (Reader out, OutputWriter writer) {
            this.out = out;
            this.writer = writer;
        }

        public void run() {
            while (true) {
                try {
                    while (!out.ready()) {
                        try {
                            Thread.currentThread().sleep(200);
                        } catch (InterruptedException e) {
                            close();
                            return;
                        }
                    }
                    if (!readOneBuffer() || Thread.currentThread().isInterrupted()) {
                        close();
                        return;
                    }
                } catch (IOException ioe) {
                    //Stream already closed, this is fine
                    return;
                }
            }
        }

        private boolean readOneBuffer() throws IOException {
            char[] cbuf = new char[255];
            int read;
            while ((read = out.read(cbuf)) != -1) {
                writer.write(cbuf, 0, read);
            }
            return read != -1;
        }

        private void close() {
            try {
                out.close();
            } catch (IOException ioe) {
                ErrorManager.getDefault().notify(ioe);
            } finally {
                writer.close();
            }
        }
    }
}
