/*
                    Sun Public License Notice

The contents of this file are subject to the Sun Public License Version
1.0 (the "License"). You may not use this file except in compliance with
the License. A copy of the License is available at http://www.sun.com/

The Original Code is the Regsup module.
The Initial Developer of the Original Code is Petr Slegr.
Portions created by Petr Slegr are Copyright (C) 2002.
All Rights Reserved.

Contributor(s): Petr Slegr.
*/

package org.netbeans.modules.regsup.ide;

import org.netbeans.modules.regsup.auted.VisualAutomatonContainer;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.IOException;
import java.awt.Point;
import org.netbeans.modules.regsup.auted.*;
import org.netbeans.modules.regsup.logic.RexEdgeElement;
import java.util.*;

/**
 *
 * @author Petr Slegr
 * @version 30.04.02
 */
public class AutomatonWriter {
    
    private VisualAutomatonContainer container;
    
    private OutputStreamWriter out;
    
    public static final String NAME_AUTOMATON = "automaton";
    
    public static final String NAME_CONTAINER = "container";
    
    public static final String NAME_NODE_ID = "nodeId";
    
    public static final String NAME_HUMAN_NAME = "humanName";
    
    public static final String NAME_LONG_HUMAN_NAME = "longHumanName";
    
    public static final String NAME_TEMPLATE_NAME = "templateName";
    
    public static final String NAME_CONTENT = "content";
    
    public static final String NAME_POSITION = "position";
    
    public static final String NAME_FINAL = "final";
    
    public static final String NAME_STARTING = "starting";
    
    public static final String NAME_X = "x";
    
    public static final String NAME_Y = "y";
    
    public static final String NAME_EDGE = "edge";
    
    public static final String NAME_NODE = "node";
    
    public static final String NAME_SOURCE = "source";
    
    public static final String NAME_TARGET = "target";
    
    public static final String NAME_VALUE = "value";
    
    private int id;
    
    /** Counter of nesting of containers
     */
    private int contCounter = -1;
    
    private static final int INDENT_BASE = 4;
    
    /** Maps from nodeId (Integer) to VisualAutomatonNode
     */
    private Hashtable nodes = new Hashtable ();
    
    /** Creates a new instance of AutomatonWriter */
    public AutomatonWriter (VisualAutomatonContainer container, OutputStream output) {
        this.container = container;
        out = new OutputStreamWriter (output);
    }

    private void writeIndent () throws IOException {
        int c = contCounter * INDENT_BASE;
        for (int i = 0; i < c; i++) {
            out.write (" ");
        }
    }
    
    /** Writes arbitrary string (content of tag) with respect to indentation
     */
    private void writeData (String str) throws IOException {
        writeIndent ();
        out.write (str);
        out.write ("\n");
    }
    
    private void writeTag (String tag, boolean close) throws IOException {
        writeIndent ();
        if (! close) {
            out.write ("<");
        } else {
            out.write ("</");
        }
        out.write (tag);
        out.write (">");
        out.write ("\n");
    }
    
    /* <!ELEMENT container (nodeId, humanName, position, final, starting, longHumanName, templateName, content)> */
    private void write (VisualAutomatonContainer container) throws IOException {
        contCounter++;
        writeTag (NAME_CONTAINER, false);
        int id1 = id++;
        nodes.put (container, new Integer (id1));
        //System.out.println("container: " + container + " with: " + (id1-1));
        writeNodeId (id1);
        writeHumanName (container.getHumanName ());
        writePosition (container.getPosition ());
        writeFinal (container.isFinalNode ());
        writeStarting (container.isStarting ());
        // !!!!!
        //writeLongHumanName (container.getLongHumanName ());
        writeLongHumanName ("");
        writeTemplateName ("");
        writeContent (container);
        writeTag (NAME_CONTAINER, true);
        contCounter--;
    }
    
    private void writeNodeId (int id) throws IOException {
        writeTag (NAME_NODE_ID, false);
        writeData (String.valueOf (id));
        writeTag (NAME_NODE_ID, true);
    }
    
    private void writeHumanName (String str) throws IOException {
        writeTag (NAME_HUMAN_NAME, false);
        writeData (str);
        writeTag (NAME_HUMAN_NAME, true);
    }
    
    private void writePosition (Point position) throws IOException {
        writeTag (NAME_POSITION, false);
        writeX (position.getX ());
        writeY (position.getY ());
        writeTag (NAME_POSITION, true);
    }
    
    private void writeFinal (boolean bo) throws IOException {
        writeTag (NAME_FINAL, false);
        writeBoolean (bo);
        writeTag (NAME_FINAL, true);
    }
    
    private void writeBoolean (boolean bo) throws IOException {
        if (bo) {
            writeData ("true");
        } else {
            writeData ("false");
        }
    }
    
    private void writeStarting (boolean bo) throws IOException {
        writeTag (NAME_STARTING, false);
        writeBoolean (bo);
        writeTag (NAME_STARTING, true);
    }
    
    private void writeLongHumanName (String str) throws IOException {
        writeTag (NAME_LONG_HUMAN_NAME, false);
        writeData (str);
        writeTag (NAME_LONG_HUMAN_NAME, true);
    }
    
    private void writeTemplateName (String str) throws IOException {
        writeTag (NAME_TEMPLATE_NAME, false);
        writeData (str);
        writeTag (NAME_TEMPLATE_NAME, true);
    }
    
    private void writeContent (VisualAutomatonContainer container) throws IOException {
        writeTag (NAME_CONTENT, false);
        Set elems = container.getElements ();
        elems.removeAll (container.getEdges ());
        for (Iterator it = elems.iterator (); it.hasNext (); ) {
            Object o = it.next ();
            if (o instanceof VisualAutomatonContainer) {
                // ! VisualAutomatonContainer must be first because is sublcass of VisualAutomatonNode
                write ((VisualAutomatonContainer) o);
            }
            else if (o instanceof VisualAutomatonNode) {
                write ((VisualAutomatonNode) o);
            }
            else if (o instanceof VisualAutomatonEdge) {
                write ((VisualAutomatonEdge) o);
            } else {
                // ... exception
            }
        }
        // ! was bug: it is necessary to store edges after all nodes (containers) are stored
        // (id assignment)
        for (Iterator it = container.getEdges ().iterator (); it.hasNext (); ) {
            Object o = it.next ();
            if (o instanceof VisualAutomatonEdge) {
                write ((VisualAutomatonEdge) o);
            } else {
                // ... exception
            }
        }
        writeTag (NAME_CONTENT, true);
    }
    
    /* <!ELEMENT node (nodeId, humanName, position, final, starting)> */
    private void write (VisualAutomatonNode node) throws IOException {
        writeTag (NAME_NODE, false);
        int id1 = id++;
        nodes.put (node, new Integer (id1));
        //System.out.println("node: " + node + " with: " + (id1-1));
        writeNodeId (id1);
        writeHumanName (node.getHumanName ());
        writePosition (node.getPosition ());
        writeFinal (node.isFinalNode ());
        writeStarting (node.isStarting ());
        writeTag (NAME_NODE, true);
    }
    
    private void writeX (double x) throws IOException {
        writeTag (NAME_X, false);
        writeData (String.valueOf ((int) x));
        writeTag (NAME_X, true);
    }
    
    private void writeY (double y) throws IOException {
        writeTag (NAME_Y, false);
        writeData (String.valueOf ((int) y));
        writeTag (NAME_Y, true);
    }
    
    /* <!ELEMENT edge (source, target, value)> */
    private void write (VisualAutomatonEdge edge) throws IOException {
        writeTag (NAME_EDGE, false);
        writeSource (edge.getSourceNode ());
        writeTarget (edge.getTargetNode ());
        writeValue (edge.getEdgeValue ());
        writeTag (NAME_EDGE, true);
    }
    
    private void writeSource (VisualAutomatonNode node) throws IOException {
        writeTag (NAME_SOURCE, false);
        Integer in = (Integer) nodes.get (node);
        if (in == null) {
            throw new IllegalStateException ("Wrong node numbering");
        }
        writeNodeId (in.intValue ());
        writeTag (NAME_SOURCE, true);
    }
    
    private void writeTarget (VisualAutomatonNode node) throws IOException {
        writeTag (NAME_TARGET, false);
        Integer in = (Integer) nodes.get (node);
        if (in == null) {
            throw new IllegalStateException ("Wrong node numbering");
        }
        writeNodeId (in.intValue ());
        writeTag (NAME_TARGET, true);
    }
    
    private void writeValue (RexEdgeElement value) throws IOException {
        writeTag (NAME_VALUE, false);
        writeData (value.toString ());
        writeTag (NAME_VALUE, true);
    }
    
    private void writeHeader () throws IOException {
        //out.write ("<!DOCTYPE automaton SYSTEM \"Automaton.dtd\">");
    }
    
    public void write () throws IOException {
        writeHeader ();
        writeTag (NAME_AUTOMATON, false);
        write (container);
        writeTag (NAME_AUTOMATON, true);
        // ! was bug - it neccessary to flush/close writer, not only stream
        out.flush ();
        out.close ();
    }
    
    
}
